package gov.nasa.gsfc.dqss;

import gov.nasa.gsfc.G4DataObject;
import gov.nasa.gsfc.Stopwatch;

import java.io.IOException;
import java.util.Iterator;
import java.util.List;

import ncsa.hdf.object.Dataset;
import ncsa.hdf.object.Group;

public class AirsMaskFactory extends MaskFactory {

	Stopwatch s;
	DataQuality dataQuality;
	
	public AirsMaskFactory(String dir, String configPath) {
		super();
		dqss = new DQSS(dir,configPath);
		DQSS.SkipBecauseNoScreeningSelected = 2;
		/*
		 * This is where the DataQuality constructor should have been in the first
		 * place. Not in startMasking to reopen the ontology each time!
		 * We will always need a DataQuality instance but not necessarily one
		 * with it's own Ontology object.
		 */
		  dataQuality = new DataQuality();  
        /*
         * If we are not using the servlet for our ontology:
         */
        if (!DQSS.UseServlet) {
        	logger.warn("Calling dataQuality...Not expecting to use Servlet for Ontology");
        	dataQuality.setup(DQSS.ontology, DQSS.navigation);
        }
	}
	/*
	 * Queries for ontology stuff in various ways so this is not needed by
	 * Masker in MODAPS Arch. Just by the Web Services prep bit in MODAPS Arch
	 */ 
	 
	public void PrepMasking(DataFile file, String snvid, String dataVariable, 
			String criteria, DQSSNumber fillvalue){
		
		
		logger.warn("HDF File:" + file.getFileName());
        logger.warn("Dataset:" + snvid);
        logger.warn("DataVariable:" + dataVariable);
        logger.warn("Criteria:" + criteria);
        logger.warn("Staging Dir:" + DQSS.stagingDirectory);
        logger.warn("FillValue:" + fillvalue);
        
		  
	       //  Ontology Servlet Architecture:
	       
	    	boolean UseCache = DQSS.UseCache;
	    	 boolean cacheTheResults = false;
	        String OntologyURL = DQSS.OntologyCriteriaURL + "?";
	        OntologyURL += "dataset=" + snvid + "&" +
	                       "parameter=" + dataVariable + "&" +
	                       "criteria=" + criteria;
	        
	        logger.debug("OntologyURL:" + OntologyURL);
	        G4DataObject obj=null;
	        String address = OntologyURL;
	        if (UseCache) {
	        	 
	        	 if (DQSS.UseServlet) {
	        		 address = OntologyURL;
	        	 }
	        	 else {
	        		 address = DQSS.ontology + "?dataset=" + snvid + "&" +
	                 "parameter=" + dataVariable + "&" +
	                 "criteria=" + criteria;
	        	 }
	        	 logger.debug("USING CACHE");
	        	 
	        	  // Cache key is set here:
	        	  
	        	 obj = dataQuality.checkCache(address);
	        	 if (obj != null) {
	        		 logger.debug("We found cache key:" + obj.getKey());
	        	 }
	        }
	        else {
	        	logger.debug("NOT USING CACHE");
	        }
	       
	        
	        // This now encapsulates Servlet and local Jena algorithms.
	         
	    	try {
	    	 dataQuality.getOntologyResults(address,UseCache,obj,snvid,dataVariable,criteria);
				
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (Exception e) {
				e.printStackTrace();
			}
			
			
			s.stop();
		    logger.debug("Get Ontology Data : elapsed time: " + s.getElapsedTime());
		    s.reset();
		    s.start();
		   
		    /*
	        int status = dataQuality.XML2Castor();
	        if (status > 0) {
	           logger.error("XML Returned from Ontology does not match what castor was expecting for dataVariable:" + dataVariable);
	           return ;
	        }
	       */
	        
	        
	         //  Put ONTOLOGY DATA:
	         //  into file object
	         
	        //dataQuality.parseAirsOntologyResults((AIRSDataFile) file);
	        
		
	}
	
	
	public  Statistics StartMasking(DataFile file,   DQSSNumber fillvalue){
		
		s = new Stopwatch();
		s.start();
	    
		
		
        if (NoScreening(file.crit2num)) {
        	return null;
        }
        
        Group newg   = file.getGroupOfDataset(dataVariable);      
	    Dataset data =   file.copyDatasetWithSuffix(dataVariable,newg, DQSS.MaskName );
	    if (false ) {
	    	String swathname = "_FV_" + dataVariable;
	    
	    	file.NavigateSwathGroup() ;
	    	Group swathg = file.getGroupOfDataset(swathname);
	    	if (swathg == null) {
	    			logger.error("Could not find group of swath:" + swathname);
	    	System.exit(3);
	    	}
	    	Dataset swath = file.copyDatasetWithSuffix(swathname,swathg, DQSS.MaskName);
	    }
	    Dataset[] quality   = file.getQSVars();
	    if (quality == null) {
	    	logger.fatal("Could not find any quality datasets for " + 
	    			dataVariable + " File " + file.getFileName() + " may not have been opened/exist");
	    	return null;
	    }
	    
	    FillValue = fillvalue;
	    DQSS.logger.debug("Using FillValue:" + fillvalue);
	    Mask mask = null;
	    if (quality != null) {
	    	logger.debug("I Found qs_var:" + quality[0].getName());
	    }
	    else {
	    	logger.error("I could not find a qs_var");
	    	return null;
	    }
	    if (data != null) {
	    	logger.debug("I Found dataVariable:" + data.getName());
	    }
	    else {
	    	logger.error("I could not find the dataVariable:" + dataVariable);
	    	return null;
	    }
	    DQSS.logger.debug(file.get_nConsRef() + ":" + file.get_nRef() + ":" + file.get_nCoordCons() + ":" + file.get_nQualLevel());
	    DQSS.logger.debug(file.nConsRef + ":" + file.nRef + ":" + file.nCoordCons + ":" + file.nQualLevel);
	    s.stop();
	    logger.debug("Get Ontology Data postprocessing: elapsed time: " + s.getElapsedTime());
	    s.reset();
	    s.start();
	    
	    /*
	     *   
  Doing TAirMWOnlyStd in Mask3D2Q with Qual_MW_Only_Temp_Strat Qual_MW_Only_Temp_Tropo nSurfStd
  - 1 CoordinateConstraint Variables and 2 QualityLevel Variables Rank 3
  
  Doing CldFrcStd     in Mask5D1Q1Con with numCloud Qual_Cloud_OLR
  - 1 CoordinateConstraint variable and 1 QualityLevel Variable  Rank 5
  
   
  Doing olr           in Mask1Q1LevelCon with numCloud Qual_Cloud_OLR
    - 1 CoordinateConstraint variable and 1 QualityLevel Variable Rank 2
  
  Doing TAirStd       in Mask3D1Q with nGoodStd
      - 1 CoordinateConstraint variable Rank 3 
      
  Doing TSurfAir      in Mask1D1Q1Ref with PBest PSurfStd
   - 1 ReferenceConstraint Variable and 1 Reference Variable Rank 1
   
  Doing totH2OMWOnlyStd in Mask1D1Q with Qual_MW_Only_H2O
   - 1 QualityLevel Variable 
  Doing TSurfStd        in Mask1D1Q with Qual_Surf
  Doing totO3Std        in Mask1D1Q with Qual_O3
  Doing CO_VMR_eff      in Mask1D1Q with Qual_CO
  Doing emisIRStd       in Mask1D1Q with Qual_Surf
  Doing totH2OStd       in Mask1D1Q with Qual_H2O
  Doing clrolr          in Mask1D1Q with Qual_clrolr
  Doing CO_total_column in Mask1D1Q with Qual_CO
  Doing CH4_VMR_eff     in Mask1D1Q with Qual_CH4
  Doing CH4_total_column in Mask1D1Q with Qual_CH4
  Doing totCldH2OStd    in Mask1D1Q with Qual_MW_Only_H2O
  
  Doing H2OMMRStd        in Mask1Q1LevelConMultiLayer  with nSurfStd Qual_H2O
  - 1 CoordinateConstraint variable and 1 QualityLevel Variable Rank 3 + minmax
  
  TT 15522
  Doing H2OMMRSat        in Mask1Q1LevelConMultiLayer  with nSurfStd Qual_H2O
  - 1 CoordinateConstraint variable and 1 QualityLevel Variable Rank 3 + minmax
  moved to: Doing H2OMMRSat(rnk:3) in Mask3D1Q with nBestStd
  
  Doing H2OMMRSat_liquid in Mask1Q1LevelConMultiLayer  with nSurfStd Qual_H2O
  moved to: Doing H2OMMRSat_liquid(rnk:3) in Mask3D1Q with nGoodStd
  - 1 CoordinateConstraint variable and 1 QualityLevel Variable Rank 3 + minmax
  
  These two conflict:
  Doing PCldTopStd    in Mask3D2Layer2QConValid with numCloud Qual_Cloud_OLR
  - 1 CoordinateConstraint variable and 1 QualityLevel Variable Rank 3 no minmax
  Doing O3VMRStd         in Mask1Q1LevelConMultiLayer  with nSurfStd Qual_O3
  - 1 CoordinateConstraint variable and 1 QualityLevel Variable Rank 3 no minmax
  
        */


		if (file.nCoordCons == 1  && file.nQualLevel == 0 && data.getRank() == 3) {
			
			logger.warn("Doing " + dataVariable + "(rnk:" + data.getRank()  + ") in Mask3D1Q with " + file.qualityNames);
			 // Doing TAirStd       in Mask3D1Q with nGoodStd
		     //  - 1 CoordinateConstraint variable Rank 3 
    	     mask = new Mask3D1Q(data, quality, file, dir);

    	} 
		else if (file.nCoordCons == 1 && file.nQualLevel == 1 &&  file.get_mins().size() == 0 ) {
			if (data.getRank() == 5) {
				logger.warn("Doing " + dataVariable + "(rnk:" + data.getRank()  + ") in Mask5D1Q1Con with " + file.qualityNames);
				// Doing CldFrcStd     in Mask5D1Q1Con with numCloud Qual_Cloud_OLR
				// - 1 CoordinateConstraint variable and 1 QualityLevel Variable  Rank 5
				mask = new Mask5D1Q1Con(data, quality, dir, file);
			}
			else if (data.getRank() == 3) {
				if (file.qs_names.contains("nSurfStd")) {
					logger.warn("Doing " + dataVariable + "(rnk:" + data.getRank()  + ") in Mask1Q1LevelConMultiLayer  with " + file.qualityNames);
					 //Doing O3VMRStd         in Mask1Q1LevelConMultiLayer  with nSurfStd Qual_O3
					 // - 1 CoordinateConstraint variable and 1 QualityLevel Variable Rank 3 no minmax
					mask = new Mask1Q1Level1ConMultiLayer(data, quality, file, dir);
				}
				else {
					// Doing PCldTopStd    in Mask3D2Layer2QConValid with numCloud Qual_Cloud_OLR
					// - 1 CoordinateConstraint variable and 1 QualityLevel Variable Rank 3 no minmax
					logger.warn("Doing " + dataVariable + "(rnk:" + data.getRank()  + ") in Mask3D2Layer2QConValid with " + file.qualityNames);
					mask = new Mask3D2Layer2QConValid(data, quality, file,  dir);
				}
			}
			else if (data.getRank() == 2) {
				logger.warn("Doing " + dataVariable + "(rnk:" + data.getRank()  + ") in Mask1Q1LevelCon with " + file.qualityNames);
				// Doing olr           in Mask1Q1LevelCon with numCloud Qual_Cloud_OLR
				// - 1 CoordinateConstraint variable and 1 QualityLevel Variable Rank 2
				// I think this won't work cuz file.get_mins().size() == 0 here
				mask = new Mask1Q1LevelCon(data, quality, file, dir);
			}
		}
//		if (dataVariable.contains("TSurfAir")) {
		
		else if (file.get_nConsRef() == 1 && file.get_nRef() == 1 ) {
			logger.warn("Doing " + dataVariable + "(rnk:" + data.getRank()  + ") in Mask1D1Q1Ref with " + file.qualityNames);
			// Doing TSurfAir      in Mask1D1Q1Ref with PBest PSurfStd
			// - 1 ReferenceConstraint Variable and 1 Reference Variable Rank 1
   	        mask = new Mask1D1Q1Ref(data, quality, file,  dir);

    	}  	   	  
//		if (dataVariable.contains("TAirMWOnly")) {
		else if (file.get_nCoordCons() == 1 && file.get_nQualLevel() == 2  && file.get_mins().size() > 0) {
			logger.warn("Doing " + dataVariable + "(rnk:" + data.getRank()  + ") in Mask3D2Q with " + file.qualityNames);
			
			 //Doing TAirMWOnlyStd in Mask3D2Q with Qual_MW_Only_Temp_Strat Qual_MW_Only_Temp_Tropo nSurfStd
			 // - 1 CoordinateConstraint Variables and 2 QualityLevel Variables Rank 3
			mask = new Mask3D2Q(data, quality,file, dir);
		}

		
		else if (file.nCoordCons == 1 && file.nQualLevel == 1 && file.get_mins().size() > 0) { 
			
			if (data.getRank() == 3) {
				logger.warn("Doing " + dataVariable + "(rnk:" + data.getRank()  + ") in Mask1Q1LevelConMultiLayer  with " + file.qualityNames);
				/*
				Doing H2OMMRStd        in Mask1Q1LevelConMultiLayer  with nSurfStd Qual_H2O
				- 1 CoordinateConstraint variable and 1 QualityLevel Variable Rank 3 + minmax
				Doing H2OMMRSat        in Mask1Q1LevelConMultiLayer  with nSurfStd Qual_H2O
				- 1 CoordinateConstraint variable and 1 QualityLevel Variable Rank 3 + minmax
				Doing H2OMMRSat_liquid in Mask1Q1LevelConMultiLayer  with nSurfStd Qual_H2O
				- 1 CoordinateConstraint variable and 1 QualityLevel Variable Rank 3 + minmax
				*/
				mask = new Mask1Q1Level1ConMultiLayer(data, quality, file, dir);
			}
			
			
		}
		
		
		else if ( file.get_nQualLevel() == 1 && file.get_nCoordCons() == 0 && 
				file.get_nRef() == 0 && file.get_nConsRef() == 0)
		 {
			  logger.warn("Doing " + dataVariable + "(rnk:" + data.getRank()  + ") in Mask1D1Q with " + file.qualityNames);
			  if (file.crit2num.size() < 1) {
				  logger.error("No User Criteria was extracted from Ontology for " +
						  data.getName());
				  return null;
			  }
			  /*
			  Doing totH2OMWOnlyStd in Mask1D1Q with Qual_MW_Only_H2O
			   - 1 QualityLevel Variable 
			  Doing TSurfStd        in Mask1D1Q with Qual_Surf
			  Doing totO3Std        in Mask1D1Q with Qual_O3
			  Doing CO_VMR_eff      in Mask1D1Q with Qual_CO
			  Doing emisIRStd       in Mask1D1Q with Qual_Surf
			  Doing totH2OStd       in Mask1D1Q with Qual_H2O
			  Doing clrolr          in Mask1D1Q with Qual_clrolr
			  Doing CO_total_column in Mask1D1Q with Qual_CO
			  Doing CH4_VMR_eff     in Mask1D1Q with Qual_CH4
			  Doing CH4_total_column in Mask1D1Q with Qual_CH4
			  */
			  mask = new Mask1D1Q(data, quality, file, dir);
		}
		
		else {
			logger.error("Could not find a case for masking for dataVariable:" + dataVariable);
			return null;
		}
		
		
		 try {
			 if (DQSS.UseReflection) {
			     mask.doMask_reflect();
			 }
			 else {
				 mask.doMask(); 
			 }
			} catch (SecurityException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (IllegalArgumentException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (NoSuchMethodException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (Throwable e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
        /*try {
				mask.doMask();
			} catch (Exception e) {
				// TODO Auto-generated catch block
				
				logger.error("We don't have a masker for: " + dataVariable + " yet");
				return null;
			}
			*/
           	// Writing stats to the global metadata now. mask.writeStats();
           
        s.stop();
        logger.debug(mask.getClass().getName() + ", " + data.getName() 
        		+ " masking elapsed time: " + s.getElapsedTime());
		return mask.getStats();
         
	}
	

	public  String LoadOntology(String snvid, String dataVariable, 
			String criteria){

		logger.warn("Dataset:" + snvid);
        logger.warn("DataVariable:" + dataVariable);
        logger.warn("Criteria:" + criteria);
        logger.warn("Staging Dir:" + DQSS.stagingDirectory);
          /*
       *  Ontology Servlet Architecture:
       */
    	boolean UseCache = DQSS.UseCache;
    	boolean cacheTheResults = false;
        String OntologyURL = DQSS.OntologyCriteriaURL + "?";
        OntologyURL += "dataset=" + snvid + "&" +
                       "parameter=" + dataVariable + "&" +
                       "criteria=" + criteria;
        
        logger.debug("OntologyURL:" + OntologyURL);
        G4DataObject obj=null;
        String address = OntologyURL;
        if (UseCache) {
        	 
        	 if (DQSS.UseServlet) {
        		 address = OntologyURL;
        	 }
        	 else {
        		 address = DQSS.ontology + "?dataset=" + snvid + "&" +
                 "parameter=" + dataVariable + "&" +
                 "criteria=" + criteria;
        	 }
        	 
        	 /*
        	  * Cache key is set here:
        	  */
        	 obj = dataQuality.checkCache(address);
        	 if (obj != null) {
        		 logger.debug("USING CACHE");
        		 logger.debug("We found cache key:" + obj.getKey());
        	 }
        	 else {
        		 logger.warn("Could not get data from cache");
        	 }
        }
        else {
        	logger.debug("NOT USING CACHE");
        }
       
        /*
         * This now encapsulates Servlet and local Jena algorithms.
         */
    	try {
    	 dataQuality.getOntologyResults(address,UseCache,obj,snvid,dataVariable,criteria);
			
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (Exception e) {
			e.printStackTrace();
		}
		arrangexml();
		
		
	    return dataQuality.results;
	}
	
    void arrangexml() {
    	String xml = dataQuality.results;
    	logger.debug("XML so far:" + xml);
    	int i = xml.indexOf("<datavar>");
        if (i < 0) {
        	
        	String errorstr = "We are now expecting each xml result to be bounded with <datavar>." +
        	" This was not found." + 
        	" We are only adding a base tag <root> after xml is returned from all datavariables."  +
        	" Perhaps old code is being referenced";
        	logger.error(errorstr);
        	System.err.println(errorstr);
        	dataQuality.results = "<datavar>invalid results</datavar>";
        }
        else {
        	dataQuality.results =   xml.substring(i, xml.length()) ;
        }
    	
    }
    
	boolean NoScreening(List<Long> criteria) {
		
		Iterator<Long> itr = criteria.iterator();
		Long skip = null;
		while (itr.hasNext()) {
			skip = itr.next();
			if (skip != DQSS.SkipBecauseNoScreeningSelected) {
				return false;
			}
		}
		if (skip == null) {
			logger.fatal("Criteria selected for this parameter was null");
			System.err.println("Criteria selected for this parameter was null" );
	        System.exit(1);
		}
		logger.warn("Skipping this parameter because no screening (" + skip + ") was selected");
		
		return true;
	}
	
	

}

