package gov.nasa.gsfc.dqss;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.util.Properties;

import ncsa.hdf.object.Dataset;
import ncsa.hdf.object.Group;

import org.apache.log4j.Logger;
import org.apache.log4j.PropertyConfigurator;

public  class DQSS {
	
	protected long rows = -1;
	protected long cols = -1;
	protected long layers = -1;
	protected String dataVariable = null;
	public static String ontology = "http://mirador.gsfc.nasa.gov/ontologies/dqss-Instances.owl";
	public static String navigation = "http://mirador.gsfc.nasa.gov/ontologies/DQSSNavigation.xml";
	protected static int  debugLevel=0;
	public static   String stagingDirectory = "/var/tmp/staging";
	public static String NoScreeningString;
	public static String MaskName;
	public static String OrigName;
	public static String RejectName;
	public static String delimiter;
	public static String Q2String0;
	public static String Q2String1;
	public static String Q2String2;
	public static String Q2String3;
	public static String Q1String0;
	public static String Q1String1;
	public static String Q1String2;
	public static String Q1String3;
	
	/*
	 * We need NoOntologyFile to signal that if UseServlet=false
	 * and we already have an ontology file, don't run DataQuality.setup();
	 */
	public static boolean NoOntologyFileProvided = false; 
	public static String OntologyCriteriaURL;
	public static boolean UseServlet = true;
	public static boolean UseCache = true;
	public static String CacheDir;
	public static float FillValue = -7777;
	public static int SkipBecauseNoScreeningSelected;
	public static String OntFile;
	public static boolean UseReflection = true;
	protected String dir;
	protected String config;
	static Logger logger = Logger.getLogger(DQSS.class.getName());

	public DQSS() {
		  /*
		   * The layer number in the SDS nBestStd is the lowest altitude 
		   * layer for which the quality is Best.
		   * if datalayer >= qualitypixel[i] set data[i] = 1, else fillvalue  
		   */
		 
	}
	public DQSS(String dir, String configPath) {
		stagingDirectory = dir;
		//log4jSession sessionLog4j = new  log4jSession(dir + File.separator, DQSS.class);
        //logger = sessionLog4j.getLogger();
        SetStagingDirectory(dir);
		ReadConfigFile(configPath);
		
		
        
	}
	
	  
	public static void SetStagingDirectory(String dir)
	{
		stagingDirectory = dir;
	}
	
	 
	 
	

	   public static void ReportError(String Message, int level) {
			if (debugLevel >= 0 && level == 0) {
				System.err.println(Message);
				DQSS.fileAppend("Serious:" + Message);
				logger.error(Message);
			}
			if (debugLevel >= 1 && level == 1) {
				//System.err.println(Message);
				DQSS.fileAppend("Warn:" + Message);
				logger.warn(Message);
			}
			if (debugLevel >= 2 && level == 2) {
				//System.out.println(Message);
				DQSS.fileAppend("Debug:" + Message);
				logger.debug(Message);
			}
			if (debugLevel >= 3 && level == 3) {
				// This is old stuff which prints without EOL char
				// that I don't need right now.
				logger.info(Message);
			}
		}
	   
	   protected static boolean fileAppend(String message) {
			 boolean status = false;
			 String stagingDir = stagingDirectory;
			    try{
			        // Create file 
			    	status = new File(stagingDir).mkdirs(); 
			    	if (!status) {
			    		File logdir = new File (stagingDir);
			    		if (!logdir.exists()) {
			    		   logger.error("Could not make " + stagingDir + " directory");
			    		   return false;
			    		}
			    	}
			    	File log = new File(stagingDir + File.separator + "masker.log");
			    	
			    	if (!log.createNewFile()) {
			    		if (!log.exists()) {
			    			logger.error("Could not create newfile: " + log.getPath());
			    			return false;
			    		}
			    	}
			    	
			        FileWriter fstream = new FileWriter(log,true);
			            BufferedWriter out = new BufferedWriter(fstream);
			        out.write(message + "\n");
			        //Close the output stream
			        out.close();
			        status = true;
			        }catch (Exception e){//Catch exception if any
			          logger.error( e.getMessage());
			        }
			        return status;
		 }
	   protected static boolean fileAppend(String path,String filename, String message) {
			 boolean status = false;
			 String stagingDir = path;
			    try{
			        // Create file 
			    	status = new File(stagingDir).mkdirs(); 
			    	if (!status) {
			    		File logdir = new File (stagingDir);
			    		if (!logdir.exists()) {
			    		   logger.error("Could not make " + stagingDir + " directory");
			    		   return false;
			    		}
			    	}
			    	File log = new File(stagingDir + File.separator + filename);
			    	
			    	if (!log.createNewFile()) {
			    		if (!log.exists()) {
			    			logger.error("Could not create newfile: " + log.getPath());
			    			return false;
			    		}
			    	}
			    	
			        FileWriter fstream = new FileWriter(log,true);
			            BufferedWriter out = new BufferedWriter(fstream);
			        out.write(message + "\n");
			        //Close the output stream
			        out.close();
			        status = true;
			        }catch (Exception e){//Catch exception if any
			          logger.error( e.getMessage());
			        }
			        return status;
		 }
	   protected static boolean fileWriteFile2BCached(String path,String filename, String message) {
           boolean status = false;
           String stagingDir = path;
              try{
                  // Create file 
                  status = new File(stagingDir).mkdirs();
                  if (!status) {
                          File logdir = new File (stagingDir);
                          if (!logdir.exists()) {
                             logger.error("Could not make " + stagingDir + " directory");
                             return false;
                          }
                  }
                  File cache = new File(stagingDir + File.separator + filename);
                  if (cache.exists()) {
                          logger.debug(cache.getName() + " exists already so not writing");
                          return true;
                  }
                  if (!cache.createNewFile()) {
                          if (!cache.exists()) {
                                  logger.error("Could not create newfile: " + cache.getPath());
                                  return false;
                          }
                  }

                  FileWriter fstream = new FileWriter(cache,false);
                      BufferedWriter out = new BufferedWriter(fstream);
                  out.write(message);
                  //Close the output stream
                  out.close();
                  status = true;
                  }catch (Exception e){//Catch exception if any
                    logger.error( e.getMessage());
                  }
                  return status;
   }

	   protected static boolean fileOverWrite(String filename, String message) {
			 boolean status = false;
			 File file = new File(filename);
			 
			 
			    try{
			        // Create file 
			    	String stagingDir = file.getParent();
			    	status = new File(stagingDir).mkdirs(); 
			    	if (!status) {
			    		File logdir = new File (stagingDir);
			    		if (!logdir.exists()) {
			    		   logger.error("Could not make " + stagingDir + " directory");
			    		   return false;
			    		}
			    	}
			    	
			    	if (!file.createNewFile()) {
			    		if (!file.exists()) {
			    			logger.error("Could not create newfile: " + file.getPath());
			    			return false;
			    		}
			    	}
			    	
			        FileWriter fstream = new FileWriter(file,false);
			            BufferedWriter out = new BufferedWriter(fstream);
			        out.write(message);
			        //Close the output stream
			        out.close();
			        status = true;
			        }catch (Exception e){//Catch exception if any
			          logger.error( e.getMessage());
			        }
			        return status;
		 }
	   public void ReadConfigFile(String configdir)
	   {
		  //System.err.println(configdir);
		  PropertyConfigurator.configure(configdir + "/dqsslog.conf");
	      String config = configdir +  "/dqss.config";
	      Properties p = new Properties();
	      File configFile = null;
	      String item = "";
	      try{
	        configFile = new File(config);
	        FileInputStream f = new FileInputStream(configFile);
	        p.load(f);
	        
	        //System.err.println("Reading:" + configFile);
	    	
	    	  item = "ONTOLOGY";
		      DQSS.ontology = p.getProperty("ONTOLOGY");
		      logger.warn("ONTOLOGY:" + ontology.trim());
		      
		      item = "NAVIGATION";
			      this.navigation = p.getProperty("NAVIGATION");
		      logger.warn("NAVIGATION:" + navigation.trim());
		      
		      item = "NOSCREEN";
		      DQSS.NoScreeningString = p.getProperty("NOSCREEN");
		      logger.warn("NOSCREEN:" + NoScreeningString.trim());
		      
		      item = "MASKNAME";
		      DQSS.MaskName = p.getProperty("MASKNAME");
		      logger.warn("MASKNAME:" + MaskName.trim());
		      
		      item = "ORIGNAME";
			    OrigName = p.getProperty("ORIGNAME");
		      logger.warn("ORIGNAME:" + OrigName.trim());
		      
		      item = "REJECTNAME";
			    RejectName = p.getProperty("REJECTNAME");
		      logger.warn("REJECTNAME:" + RejectName.trim());
		      
		      item = "ONTOLOGYURL";
		      DQSS.OntologyCriteriaURL = p.getProperty("ONTOLOGYURL");
		      logger.warn("ONTOLOGYURL:" + OntologyCriteriaURL.trim());
	
		      item = "USECACHE";
		      DQSS.UseCache = new Boolean(p.getProperty("USECACHE"));
		      logger.warn("USECACHE:" + UseCache);
		      
		      item = "USESERVLET";
		      DQSS.UseServlet = new Boolean(p.getProperty("USESERVLET"));
		      logger.warn("USESERVLET:" + UseServlet);
		      
		      item = "CACHEDIR";
		      DQSS.CacheDir = p.getProperty("CACHEDIR");
		      logger.warn("CACHEDIR:" + CacheDir.trim());
		      
		      item = "USEREFLECTION";
		      DQSS.UseReflection = new Boolean(p.getProperty("USEREFLECTION"));
		      logger.warn("USEREFLECTION:" + UseReflection);
		      
		      item = "DELIMITER";
		      DQSS.delimiter = p.getProperty("DELIMITER");
		      logger.warn("DELIMITER:" + delimiter.trim());
		      
		      item = "Q1STRING0";
		      DQSS.Q1String0 = p.getProperty("Q1STRING0");
		      logger.warn("Q1STRING0:" + Q1String0.trim());
		      item = "Q1STRING1";
		      DQSS.Q1String1 = p.getProperty("Q1STRING1");
		      logger.warn("Q1STRING1:" + Q1String1.trim());
		      item = "Q1STRING2";
		      DQSS.Q1String2 = p.getProperty("Q1STRING2");
		      logger.warn("Q1STRING2:" + Q1String2.trim());
		      item = "Q1STRING3";
		      DQSS.Q1String3 = p.getProperty("Q1STRING3");
		      logger.warn("Q1STRING3:" + Q1String3.trim());
		      
		      item = "Q2STRING0";
		      DQSS.Q2String0 = p.getProperty("Q2STRING0");
		      logger.warn("Q2STRING0:" + Q2String0.trim());
		      item = "Q2STRING1";
		      DQSS.Q2String1 = p.getProperty("Q2STRING1");
		      logger.warn("Q2STRING1:" + Q2String1.trim());
		      item = "Q2STRING2";
		      DQSS.Q2String2 = p.getProperty("Q2STRING2");
		      logger.warn("Q2STRING2:" + Q2String2.trim());
		      item = "Q2STRING3";
		      DQSS.Q2String3 = p.getProperty("Q2STRING3");
		      logger.warn("Q2STRING3:" + Q2String3.trim());
		      /* Decided not to do this at all
		       * 
		      
		      item = "SKIPVALUE";
		      String SkipValue= p.getProperty("SKIPVALUE");
		      logger.warn("SKIPVALUE:" + SkipValue.trim());
		      SkipBecauseNoScreeningSelected = Integer.parseInt(SkipValue);
		       */
	      }
	      catch(Exception e){
	    	  logger.error("Error loading config file: " + configFile + 
	    			  " perhaps " + item + " is missing from config" );
	    	  System.err.println("Error loading config file: " + configFile +
	    			  " perhaps " + item + " is missing from config" );
	    	  
		      System.exit(1);
	      }

	      
	   }
	   public  static float getFillValue() {
			return FillValue;
		}
}
