package gov.nasa.gsfc.dqss;

import java.lang.reflect.Method;

import ncsa.hdf.object.Dataset;

public class Mask1D1Q extends Mask {
	String direction = null;
	
	public Mask1D1Q(Dataset data, Dataset[] qual, DataFile File , String dir) {
		super(data,qual,dir,File);
		direction = File.qs_direction.get(0);	
	}

	
	public void doMask() throws Exception {
		/*
    	 * Ok, what I am going to do here is to read in TAirStd and nBestStd.
    	 * The Dataset.read() method should read nBestStd in as one array. 
    	 * Whether it is in 2D or not I don't care as long as it reads in each
    	 * TAirStd Layer in the same manner.
    	 * So I will read the first layer.
    	 * For every qs[i] if the layer number < qs[i] I'll set it to 0 otherwise
    	 * I'll set it to 1. 
    	 */
    
    	
   
    	short[] qualityData = (short[]) quality[0].read();
    	float fill = (float) MaskFactory.getFillValue().floatValue();
    	long[] start = data.getStartDims();

    	for (int layer = 0; layer < layers ; ++layer) {
    		if (start.length > 2) {
        		start[2] = layer;
        	}	
    		logger.debug("Masking Layer:" + layer);
    		
    	    float[] dataData = (float[]) data.read();
    	    
    	    if (direction.contains("negative") ) {
        		for (int i = 0; i < rows * cols; ++i) {
        			if (dataData[i] == fill) {
        				stats.countfill();
        			}
        			if (qualityData[i] <= pCrit) {	    		
        				dataData[i] = 1;
        				stats.inc();
        			}
        			else {
        				dataData[i] = fill;
        			}
        		}
    		 } // directions
    	     else if (direction.contains("positive") ) {
    	    		for (int i = 0; i < rows * cols; ++i) {
    	    			if (dataData[i] == fill) {
    	    				stats.countfill();
    	    			}
    	    			if (qualityData[i] >= pCrit) {	    		
    	    				dataData[i] = 1;
    	    				stats.inc();
    	    			}
    	    			else {
    	    				dataData[i] = fill;
    	    			}
    		    
    		 } // just two in this case.
    		    
    		}
    		data.write(dataData);
	   }
    	stats.setTotal((int)(layers * rows * cols));
    		
    		
    	
		
	}

	/*
	 * Reflection version
	 * @see gov.nasa.gsfc.dqss.Mask#doMask()
	 */
	public void doMask_reflect() throws Exception {
		/*
    	 * Ok, what I am going to do here is to read in TAirStd and nBestStd.
    	 * The Dataset.read() method should read nBestStd in as one array. 
    	 * Whether it is in 2D or not I don't care as long as it reads in each
    	 * TAirStd Layer in the same manner.
    	 * So I will read the first layer.
    	 * For every qs[i] if the layer number < qs[i] I'll set it to 0 otherwise
    	 * I'll set it to 1. 
    	 */
		 ReflectDirections rf = new ReflectDirections();
	     Method refl = rf.getMethod(direction,Short.TYPE, Long.TYPE);
	     Object reflist[] = new Object[2];
	   
    	
   
    	short[] qualityData = (short[]) quality[0].read();
    
    	//Number fill =  DataFile.getFillValueN(data);
    	float fill = (float) MaskFactory.getFillValue().floatValue();
    	long[] start = data.getStartDims();
    	logger.debug("length:" + qualityData.length);
    	logger.debug("rows * cols:" + rows * cols );
    	
        reflist[1] = pCrit;
    	for (int layer = 0; layer < layers ; ++layer) {
    		if (start.length > 2) {
        		start[2] = layer;
        	}	
    		logger.debug("Masking Layer:" + layer);
    		
    	    float[] dataData = (float[]) data.read();
    		for (int i = 0; i < rows * cols; ++i) {
    			reflist[0] = qualityData[i];
    			if (dataData[i] == fill) {
    				stats.countfill();
    			}
    			if ( (Boolean) refl.invoke(rf, reflist)) {
    			//if (qualityData[i] <= pCrit) {	    		
    				dataData[i] = 1;
    				stats.inc();
    			}
    			else {
    				dataData[i] = fill;
    				
    			}
    		    
    		}
    		data.write(dataData);
	   }
    	stats.setTotal((int)(layers * rows * cols));
    		
    		
    	
		
	}

}
