package gov.nasa.gsfc.dqss;

import java.lang.reflect.Method;
import java.util.HashMap;

import ncsa.hdf.object.Dataset;

/*
 * This class compares the layer of the data with a coordinate constraint variable
 * Then if it is > than the coordcons variable (in this case nSurfStd))
 * it looks at the the minmax of 2 QualityLevelVariables.
 * Whichever QualityLevelVariable it is between it then
 * compares it against the user given criteria
 */

/*
 * <datavar><!--XML result from SPARQL Query-->
        <queryOutput>
                <QualityView>TAirMWOnlyStdView</QualityView>
                        <ScreeningVariable>nSurfStd</ScreeningVariable>
                        <ScreenVarType>CoordinateConstraintVariable</ScreenVarType>
                        <dim>StdPressureLev</dim>
                        <min/><max/>
                        <direction/><Level/>
                        <constraint_relationship>lessThan</constraint_relationship>
        </queryOutput>
        <queryOutput>
                <QualityView>TAirMWOnlyStdView</QualityView>
                        <ScreeningVariable>Qual_MW_Only_Temp_Strat</ScreeningVariable>
                        <ScreenVarType>QualityLevelVariable</ScreenVarType>
                        <dim>StdPressureLev</dim>
                        <min>11</min>
                        <max>28</max>
                        <direction>negative</direction>
                        <Level>0</Level>
        </queryOutput>
        <queryOutput>
        <QualityView>TAirMWOnlyStdView</QualityView>
                <ScreeningVariable>Qual_MW_Only_Temp_Tropo</ScreeningVariable>
                <ScreenVarType>QualityLevelVariable</ScreenVarType>
                <dim>StdPressureLev</dim>
                <min>1</min>
                <max>10</max>
                <direction>negative</direction>
                <Level>0</Level>
        </queryOutput>
</datavar>

 */
public class Mask3D2Q extends Mask {
	
	 
	DataFile file;  
	short[] qualData1 = null;
	short[] qualData2 = null;
	int[] coordCon  = null;
	long[] qual1Limits = null;
    long[] qual2Limits = null;
    String[] direction = null;
    float[] dataData = null;
    float fill;
    
	public Mask3D2Q(Dataset data, Dataset[] quality, DataFile File, String dir) {
		super(data,quality,dir,File);
		file = File;
		qual1Limits = new long[file.mins.size()];
	    qual2Limits = new long[file.mins.size()];
	    direction = new String[file.mins.size()];
	}
	/*
	ReflectDirections rf = new ReflectDirections();
	Class c = null;
	try {
		c = Class.forName("gov.nasa.gsfc.dqss.ReflectDirections");
	} catch (ClassNotFoundException e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
	}
	 Method m[] = c.getDeclaredMethods();
	 Class partypes[] = new Class[2];
     partypes[0] = Integer.TYPE;
     partypes[1] = Integer.TYPE;
       
	 Method meth = c.getMethod("lt", partypes);
	 
	   s.reset();
	 s.start();
	 int temp = 0;
	 boolean retval;
	 Object arglist[] = new Object[2];
	 for (int pixel = 0 ; pixel < 250000; ++pixel ) {
     	 
    	 arglist[0] = pixel;
    	 arglist[1] = 9980;
		 Object resp = meth.invoke(rf,arglist);
	     retval = (Boolean)resp;
		 if (retval) {
			 temp = pixel;
				   //System.out.println(pixel);
		 }
		   
	  }
      System.out.println("Elapsed Time in millisecs: " + s.getElapsedTime());
	*/
   

    /*
     * Reflection version:
     */
    public void doMask_reflect() throws Exception
    {
   /* Tropo is 1-10
    * Strat is 11 -28
    * nSurfStd is base level. 
    */
    
   
    
    logger.debug(data.getName() + " Min DataSlice size:" + file.mins.size());
    if (file.mins.get(0)  < file.mins.get(1)) {
    	qual1Limits[0] = file.get_mins().get(0);
    	qual2Limits[0] = file.get_mins().get(1);
    	qual1Limits[1] = file.get_maxs().get(0);
    	qual2Limits[1] = file.get_maxs().get(1);
    	direction[0]   = file.get_direction().get(0);
    	direction[1]   = file.get_direction().get(1);
    		
    }
    else {
    	qual1Limits[0] = file.mins.get(1);
    	qual2Limits[0] = file.mins.get(0);
    	qual1Limits[1] = file.maxs.get(1);
    	qual2Limits[1] = file.maxs.get(0);
    	direction[0]   = file.get_direction().get(1);
    	direction[1]   = file.get_direction().get(0);
    }
	logger.debug("Tropo Limits:" + qual1Limits[0] + " - " + qual1Limits[1]);
	logger.debug("Strat Limits:" + qual2Limits[0] + " - " + qual2Limits[1]);
	
   // HashMap q1 = new HashMap();  
    //DataFile.determineDatatype(quality[0],q1);
    
  
    for (int i = 0; i < quality.length; ++i) {
    	logger.debug("Quality[" + i + "]" + quality[i].getName());
    	if (file.get_qs_types().get(i).contains("QualityLevelVariable")) {
    		if (quality[i].getName().contains("Tropo")) {
    		    qualData1 = (short[]) quality[i].read();
    		    logger.debug("Quality 1:" + quality[i].getName());
    		}
    		else if (file.get_qs_types().get(i).contains("QualityLevelVariable")) {
    			qualData2 = (short[]) quality[i].read();
    			logger.debug("Quality 2:" + quality[i].getName());
    		}
    	}
    	else {
    		coordCon = (int[]) quality[i].read();
    	}
    }
	
	
	long[] start = data.getStartDims(); // get startDims so we can navigate the layers
	long nlayer = start[2];
	float fill = (float) MaskFactory.getFillValue().floatValue();
	
	logger.debug ("length:" + qualData1.length);
	logger.debug ("Fill value:" + fill);

	/*
	 * Determine which reflection methods to use
	 */
	logger.debug("directions:" + direction[0] + "," + 
			   direction[1] + "," + file.qs_coordcon_dir.get(0));
	 ReflectDirections rf = new ReflectDirections();
	 //CC
     Method cc = rf.getMethod(file.qs_coordcon_dir.get(0),Integer.TYPE, Integer.TYPE);
     Object cclist[] = new Object[2];
     //quality1 direction
     Method q1 = rf.getMethod(direction[0],Short.TYPE, Long.TYPE);
     Object q1list[] = new Object[2];
     //quality2 direction
     Method q2 = rf.getMethod(direction[1],Short.TYPE, Long.TYPE);
	 Object q2list[] = new Object[2];

   
	  for (int layer = 0 ; layer < layers; ++layer) {
		start[2] = layer; // this sets the layer outside the class as dims are passed by reference.
		dataData = (float[]) data.read();

		   q1list[1] = pCrit;
		   q2list[1] = pCrit;
		   cclist[1] = layer + indexOffset;
		   for (int i = 0; i < rows * cols; ++i) {
			   
			   cclist[0] = coordCon[i];
			   q1list[0] = qualData1[i];
			   q2list[0] = qualData2[i];
            
			   if (dataData[i] == fill) {
                      stats.countfill();
               }
               /*
                * If layer >= nSurfStd
                */
               else if ((Boolean)cc.invoke(rf, cclist)) {
                   if (layer + indexOffset >= qual1Limits[0] && 
                	   layer + indexOffset <= qual1Limits[1]) {
                        /*
                         * We compare with Tropo 
                         */
                         if ( (Boolean) q1.invoke(rf, q1list)) {
                                  dataData[i] = 1;
                                   stats.inc();
                         }
                         else {
                               dataData[i] = fill;
                         }
                   }
                   else {
                      /*
                       * We compare against Strat
                       */
                       if ((Boolean) q2.invoke(rf, q2list)) {
                           dataData[i] = 1;
                           stats.inc();
                       }
                       else {
                           dataData[i] = fill;
                       }
                   }
               }
               else {
                           dataData[i] = fill;
               }
			   //logger.debug("CC:" + coordCon[i] + " q1:" + qualData1[i] + " q2:" + qualData2[i]  + " result:" + dataData[i] + " pCrit:" + pCrit);

          }

     
		data.write(dataData);
		
	}
	stats.setTotal((int) (layers * rows * cols));
	
}

	  public void doMask() throws Exception
	    {
		   /* Tropo is 1-10
		    * Strat is 11 -28
		    * nSurfStd is base level. 
		    */
		    
		   
		    
		    logger.debug(data.getName() + " Min DataSlice size:" + file.mins.size());
		    if (file.mins.get(0)  < file.mins.get(1)) {
		    	qual1Limits[0] = file.get_mins().get(0);
		    	qual2Limits[0] = file.get_mins().get(1);
		    	qual1Limits[1] = file.get_maxs().get(0);
		    	qual2Limits[1] = file.get_maxs().get(1);
		    	direction[0]   = file.get_direction().get(0);
		    	direction[1]   = file.get_direction().get(1);
		    		
		    }
		    else {
		    	qual1Limits[0] = file.mins.get(1);
		    	qual2Limits[0] = file.mins.get(0);
		    	qual1Limits[1] = file.maxs.get(1);
		    	qual2Limits[1] = file.maxs.get(0);
		    	direction[0]   = file.get_direction().get(1);
		    	direction[1]   = file.get_direction().get(0);
		    }
	    	logger.debug("Tropo Limits:" + qual1Limits[0] + " - " + qual1Limits[1]);
	    	logger.debug("Strat Limits:" + qual2Limits[0] + " - " + qual2Limits[1]);
	    	
		    HashMap q1 = new HashMap();  
            DataFile.determineDatatype(quality[0],q1);
            for (int i = 0; i < quality.length; ++i) {
            	logger.debug("Quality[" + i + "]" + quality[i].getName());
            	if (file.get_qs_types().get(i).contains("QualityLevelVariable")) {
            		if (quality[i].getName().contains("Tropo")) {
            		    qualData1 = (short[]) quality[i].read();
            		    logger.debug("Quality 1:" + quality[i].getName());
            		}
            		else if (file.get_qs_types().get(i).contains("QualityLevelVariable")) {
            			qualData2 = (short[]) quality[i].read();
            			logger.debug("Quality 2:" + quality[i].getName());
            		}
            	}
            	else {
            		coordCon = (int[]) quality[i].read();
            	}
            }
	    	
	    	
	    	long[] start = data.getStartDims(); // get startDims so we can navigate the layers
	    	long nlayer = start[2];
	    	float fill = (float) MaskFactory.getFillValue().floatValue();
	    	
	    	logger.debug ("length:" + qualData1.length);
	    	logger.debug ("Fill value:" + fill);
	    
	    	for (int layer = 0 ; layer < layers; ++layer) {
	    		start[2] = layer; // this sets the layer outside the class as dims are passed by reference.
	    		dataData = (float[]) data.read();
	    		

	    			   for (int i = 0; i < rows * cols; ++i) {

                           if (dataData[i] == fill) {
                                   stats.countfill();
                           }
                           /*
                            * If layer >= nSurfStd
                            */
                           else if (layer + indexOffset >= coordCon[i]) {
                                   if (layer + indexOffset >= qual1Limits[0] && layer + indexOffset <= qual1Limits[1]) {
                                           /*
                                            * We compare with Tropo 
                                            */
                                           if ( qualData1[i] <= pCrit) {

                                                          dataData[i] = 1;
                                                          stats.inc();
                                                  }
                                                  else {
                                                          dataData[i] = fill;
                                                  }

                                   }
                                   else {
                                           /*
                                            * We compare against Strat
                                            */
                                           if (qualData2[i] <= pCrit) {

                                                          dataData[i] = 1;
                                                          stats.inc();
                                                  }
                                                  else {
                                                          dataData[i] = fill;
                                                  }


                              }
                           }
                           else {
                                   dataData[i] = fill;
                           }


                  }
	    	      data.write(dataData);
	    		}
	    		
	    		
	    	
	    	stats.setTotal((int) (layers * rows * cols));
	    	
	    }
	  
	  public void doMask_silly() throws Exception
	    {
		
		    logger.debug(data.getName() + " Min DataSlice size:" + file.mins.size());
		    if (file.mins.get(0)  < file.mins.get(1)) {
		    	qual1Limits[0] = file.get_mins().get(0);
		    	qual2Limits[0] = file.get_mins().get(1);
		    	qual1Limits[1] = file.get_maxs().get(0);
		    	qual2Limits[1] = file.get_maxs().get(1);
		    	direction[0]   = file.get_direction().get(0);
		    	direction[1]   = file.get_direction().get(1);
		    		
		    }
		    else {
		    	qual1Limits[0] = file.mins.get(1);
		    	qual2Limits[0] = file.mins.get(0);
		    	qual1Limits[1] = file.maxs.get(1);
		    	qual2Limits[1] = file.maxs.get(0);
		    	direction[0]   = file.get_direction().get(1);
		    	direction[1]   = file.get_direction().get(0);
		    }
		    //String[] ccrel = new String[file.qs_coordcon_dir.size()];
		    
	    	logger.debug("Tropo Limits:" + qual1Limits[0] + " - " + qual1Limits[1]);
	    	logger.debug("Strat Limits:" + qual2Limits[0] + " - " + qual2Limits[1]);
	    	
		    HashMap q1 = new HashMap();  
          DataFile.determineDatatype(quality[0],q1);
          for (int i = 0; i < quality.length; ++i) {
          	logger.debug("Quality[" + i + "]" + quality[i].getName());
          	if (file.get_qs_types().get(i).contains("QualityLevelVariable")) {
          		if (quality[i].getName().contains("Tropo")) {
          		    qualData1 = (short[]) quality[i].read();
          		    logger.debug("Quality 1:" + quality[i].getName());
          		}
          		else if (file.get_qs_types().get(i).contains("QualityLevelVariable")) {
          			qualData2 = (short[]) quality[i].read();
          			logger.debug("Quality 2:" + quality[i].getName());
          		}
          	}
          	else {
          		coordCon = (int[]) quality[i].read();
          	}
          }
	    	
	    	
	    	long[] start = data.getStartDims(); // get startDims so we can navigate the layers
	    	long nlayer = start[2];
	    	fill = (float) MaskFactory.getFillValue().floatValue();
	    	
	        
	    	for (int layer = 0 ; layer < layers; ++layer) {
	    		start[2] = layer; // this sets the layer outside the class as dims are passed by reference.
	    		dataData = (float[]) data.read();
	    		 logger.debug("directions:" + direction[0] + "," + 
  					   direction[1] + "," + file.qs_coordcon_dir.get(0));
	    		   if (direction[0].contains("negative") &&
	    			   direction[1].contains("negative") &&
	    			   file.qs_coordcon_dir.get(0).contains("lessThan")) {
	    			  
	    			   lessthan_neg_neg(layer);
	    		   } // directions
	    		   else if (direction[0].contains("positive") &&
		    			   direction[1].contains("positive") &&
		    			   file.qs_coordcon_dir.get(0).contains("lessThan")) {
		    			  
		    			   lessthan_pos_pos(layer);
		    		   } // directions
	    		   else if (direction[0].contains("negative") &&
		    			   direction[1].contains("negative") &&
		    			   file.qs_coordcon_dir.get(0).contains("equals")) {
		    			  
		    			  // equal_neg_neg(layer);
		    		   } // directions
	    		   else if (direction[0].contains("positive") &&
		    			   direction[1].contains("positive") &&
		    			   file.qs_coordcon_dir.get(0).contains("equals")) {
		    			  
		    			  // equals_pos_pos(layer);
		    		   } // directions
	    		   else if (direction[0].contains("positive") &&
		    			   direction[1].contains("positive") &&
		    			   file.qs_coordcon_dir.get(0).contains("lessThan")) {
		    			  
		    			   lessthan_pos_pos(layer);
		    		   } // directions
	    		    data.write(dataData);
	    	} // foreach layer
			   
	    	stats.setTotal((int) (layers * rows * cols));
	    }
	  
	    private void lessthan_neg_neg(long layer) 
	    {
		  for (int i = 0; i < rows * cols; ++i) {
              if (dataData[i] == fill) {
                      stats.countfill();
              }
              else if (coordCon[i] <= layer + indexOffset ) {
                      if (layer + indexOffset >= qual1Limits[0] && layer + indexOffset <= qual1Limits[1]) {
                              if ( qualData1[i] <= pCrit) {
                                             dataData[i] = 1; stats.inc();
                                     }
                                     else {  dataData[i] = fill; 
                              }
                      }
                      else {
                     	     if (qualData2[i] <= pCrit) {
                                             dataData[i] = 1;    stats.inc();
                                     }
                                     else {    dataData[i] = fill;
                                     }
                      }
              }
           } // foreach pixel

	  }

	    
	    private void lessthan_pos_pos(long layer ) 
	    {
		  for (int i = 0; i < rows * cols; ++i) {
              if (dataData[i] == fill) {
                      stats.countfill();
              }
              else if (coordCon[i] <= layer + indexOffset ) {
                      if (layer + indexOffset >= qual1Limits[0] && layer + indexOffset <= qual1Limits[1]) {
                              if ( qualData1[i] >= pCrit) {
                                             dataData[i] = 1; stats.inc();
                                     }
                                     else {  dataData[i] = fill; 
                              }
                      }
                      else {
                     	     if (qualData2[i] >= pCrit) {
                                             dataData[i] = 1;    stats.inc();
                                     }
                                     else {    dataData[i] = fill;
                                     }
                      }
              }
           } // foreach pixel

	  }

}


