package gov.nasa.gsfc.dqss;

import java.lang.reflect.Method;

import ncsa.hdf.object.Dataset;

/* MOD05 IR  - so not using yet */
public class Modis1D1Q extends Mask {

	/*
	 * This class won't take care of both the IR and the NIR because the
	 * Cloud Mask is not needed for the IR. This is good because the Cloud_Mask_QA
	 * is 2030x1354 whereas IR is that divided by 5 or 406x270
	 * So that's how we can remember that CMQA is not needed for IR.
	 * Chris has modeled in his XML stub that the 
	 * IR QA is Total Precip Water Confidence Flag and the 
	 * MODIS Atmosphere QA Plan shows that the this info is in the first byte of the 
	 * Quality_Assurance_Infrared
	 * Chris has modeled that NIR will have quality variables 
	 * Quality_Assurance_Near_Infrared/Total Precip Water Confidence Flag 
	 * AND Cloud_Mask_QA/Cloud Mask Cloudiness flag. 
	 * The MODIS QA doc shows that the Total Precip Conf flag is in nits 2,3,4
	 * and the Cloudiness Flag is in bits 2,3. 
	 */
	
	 String direction = null;
	 DataFile file = null;
	public Modis1D1Q(Dataset data, Dataset[] qual, DataFile File , String dir) {
		super(data,qual,dir,File);
		file = File;
		if  (file.get_direction() != null) {
			direction = file.get_direction().get(0);
		}
		else {
			logger.error("No direction obtained for " + data.getName());
		}
	}

	 public void doMask_reflect() throws Exception
	  {
		 /*
		  * MOD05_L2 is so fast we can live with reflection
		  */
		  doMask();
	  }
	public void doMask() throws Exception {
		/*
    	 * Ok, what I am going to do here is to read in TAirStd and nBestStd.
    	 * The Dataset.read() method should read nBestStd in as one array. 
    	 * Whether it is in 2D or not I don't care as long as it reads in each
    	 * TAirStd Layer in the same manner.
    	 * So I will read the first layer.
    	 * For every qs[i] if the layer number < qs[i] I'll set it to 0 otherwise
    	 * I'll set it to 1. 
    	 */
    
		ReflectDirections rf = new ReflectDirections();
	   	 
        Method qa = rf.getMethod(direction,Integer.TYPE, Long.TYPE);
        Object qalist[] = new Object[2];
        Object usefullist[] = new Object[2];
        long[] pCrit = new long[file.crit2num.size()] ;
    	for (int i = 0; i < pCrit.length; ++i) {
    		pCrit[i] = file.crit2num.get(i);
    	}
    	
    	qalist[1] = pCrit[0];
    	usefullist[1] = 0;
    	
   
    	byte[] qualityData = (byte[]) quality[0].read();
    	//Number fill =  DataFile.getFillValueN(data);
    	short fill = (short) MaskFactory.getFillValue().shortValue();
    	long[] start = data.getStartDims();
    	logger.debug("length:" + qualityData.length);
    	logger.debug("rows * cols:" + rows * cols );
    	float before;
    
    	
    	
    	for (int layer = 0; layer < layers ; ++layer) {
    		if (start.length > 2) {
        		start[2] = layer;
        	}	
    		logger.debug("Masking Layer:" + layer);
    		
    	    short[] dataData = (short[]) data.read();
    	    logger.debug("length: data:" + dataData.length);
    	    stats.setpixels(rows, cols);
    	    
    	    for (int i = 0; i < rows * cols; ++i) {
    	    	//Suraiya said not to use usefullness.
    	       // usefullist [0] = evaluateUsefulnessBits(qualityData[i]);
    	    	qalist[0] = evaluateTotalPrecipConfFlag(qualityData[i]);
    	    	//stats.countQAflags(evaluateTotalPrecipConfFlag(qualityData[i]));
    	    	/* This test showed only case where qa was 3 and data was fill
    		    if ((qualityData[i] == 0 && dataData[i] != fill) ||
    		        (qualityData[i] > 1  && dataData[i] == fill )  ) {
    		        	logger.debug("qa:" + qualityData[i] + " data:" + dataData[i]);
    		    }
    		    */
    	    	if (dataData[i] == fill) {
    	    		//logger.debug("filling");
    	    		dataData[i] = fill;
    	    		stats.countfill();
    	    	}
    	    	else if (/*(Boolean) qa.invoke(rf,usefullist) && */
    	    			 (Boolean) qa.invoke(rf , qalist) )
    	       { 
    	    		dataData[i] = 1;
    	    		stats.inc();
    	    		//logger.debug("asserting mask pixel");
    	    	}
    	    	else {
    	    		dataData[i] = fill;
    	    	}
    			
    			
    		    
    		}
    		data.write(dataData);
	   }
    	stats.setTotal((int)(layers * rows * cols));
        //stats.publishQAFlags();		
    		
    	
		
	}
	/* Java internally uses (or emulates) twos complement arithmetic. 
	 * Positive numbers are stored in binary. That means that for example the number 13 would be written 
	 * 0x0d = 00001101 = 1*8 + 1*4 + 0*2 + 1*1. 
	 * The high order bit is called the sign bit. For signed quantities such as byte, short, int and long, 
	 * when that high order bit is a 1, we consider the number negative. 
	 * To represent a negative number we first invert each bit, then add 1. 
	 * We would write -13 this way: 0x0d -  00001101 -> 11110010 -> 11110011 -> 0xf3.
	 * 
	 * two's complement of a number is obtained by taking negative (-num)
	 * 
	 * If number is negative then bit 7 (see MODAPS Atmosphere QA Plan for 005)
	 * is 1. To get the rest of the numbers take the negative and and it.
	 */
	
	/* The Water Vapor (or Total Precipitable Water) product combines results from
       both the NIR (1x1 km) and IR algorithms (5 x 5 km). For near infrared (NIR) total
       precipitable water, the first byte contains cloud mask QA (1 x 1 km), and the second
       byte (a separate array from the cloud mask QA) contains NIR product quality and
       retrieval processing flags. 
       For the five bytes of IR total precipitable water results, only
       product quality and retrieval processing flags are stored. Since IR total precipitable
       water results are copied from 07_L2, the cloud mask related QA flags can be retrieved
       from 07_L2, and therefore it will not be duplicated here in the 05_L2 product. 
    */
	/* Quality_Assurance_Near_Infrared: Length: 1 byte (8 bits)
	   Flag Name						Number of Bits		Bit Values		Definitions
	   Total Precipitable Water (NIR)	1					0,1				Not useful, Useful
			   Usefulness flag
		Total Precipitable Water (NIRT) 3					0,1,2,3,		Bad(Fill) Marginal,Good,VG
		Confidence Flag
		Inversion Method Used (NIR)     2                   0,1,2           2 chan ratio, 3 chan ratio, no retrival
		Surface Type                    2					0,1,2,3			Bright Land, Clear Ocean, Cloud, Sunglint
		
	*/
	private boolean evaluateQualityAssuranceNearInfraredBits(Byte a) {
		//byte[] bits = {1,2,4,8,16,32,64,-128};
		byte[] bits = {-128,64,32,16,8,4,2,1};
		byte usefulnessMask = 1;
		byte surfaceMask    = -64 ; // and the sign of the byte.
		byte confidenceMask = 2 + 4 + 8;
		byte test = 32 + 64;
		byte alwayszeroMask = 8;
		Integer b = a & 0xFF;
		
		
		int surface = a & surfaceMask;
		if (surface == 64) {
			surface = 1;
		}
		else if (surface == -128) {
			surface = 2;
		}
		else if (surface == -64) {
			surface = 3;
		}
		
		int usefulness = a & usefulnessMask;
		
		/* Confidence */
		int confidence = a & confidenceMask;
		if (confidence == 6) {
			confidence = 3;
		}
		else if (confidence == 2) {
			confidence = 1;
		}
		else if (confidence == 4) {
			confidence = 2;
		}
		else  if (confidence == 0) {
			
		}
		else {
			confidence = 9999;
		}
		if (confidence > 1 && usefulness == 1) {
			return true;
		}
		return false;
}

	private void evaluateBits(Byte a, int count) {
		//byte[] bits = {1,2,4,8,16,32,64,-128};
		byte[] bits = {-128,64,32,16,8,4,2,1};
		byte usefulnessMask = 1;
		byte surfaceMask    = -64 ; // and the sign of the byte.
		byte confidenceMask = 2 + 4 + 8;
		byte test = 32 + 64;
		byte alwayszeroMask = 8;
		Integer b = a & 0xFF;
		int line = count / 270;
		int pixel = count - 270 * line;
		int surface = a & surfaceMask;
		if (surface == 64) {
			surface = 1;
		}
		else if (surface == -128) {
			surface = 2;
		}
		else if (surface == -64) {
			surface = 3;
		}
		//System.out.print(a  + " ->  " + b + " = " );
		/* alwayszero test 4th bit is not used as confidence Flag is 0-3 but uses 3 bits*/
		int alwayszero = count & alwayszeroMask;
		if (alwayszero == 0) {
			//System.out.print (count + ",");
		}
		else {
			//System.out.print (" ,");
		}
		/* Usefulness: */
		int usefulness = a & usefulnessMask;
		//System.out.println ("Usefulness:" + usefulness);
		
		/* Confidence */
		int confidence = a & confidenceMask;
		if (confidence == 6) {
			confidence = 3;
		}
		else if (confidence == 2) {
			confidence = 1;
		}
		else if (confidence == 4) {
			confidence = 2;
		}
		else  if (confidence == 0) {
			
		}
		else {
			confidence = 9999;
		}
		System.out.println ("Confidence:" + confidence + " Usefulness:" + usefulness + 
				" surface:" + surface + " cnt:" + count + " :" + line + ":" + pixel);
		for (int i = 0; i < 8; ++i) {
			Integer value = bits[i] & confidenceMask; // this gives the twos complement
			//System.out.print(value  + ", ");
			
		}
		/* this gives the two's complement of the stored number which is already the two's complement
		 * of the actual number so it's what we want -- except -- signed bit... You get value of this
		 * bit by whether number is positive or negative.
		 */
		if (confidence == 9999 || surface == 3) {
		for (int i = 0; i < 8; ++i) {
			if (i == 0 ) {
				System.out.print(a + ":");
			}
			Integer value = bits[i] & a; // this gives the twos complement
			System.out.print(value  + ", ");
			if (i == 7 ) {
				System.out.println(" - " + count);		
			}
		}
		}
	}
	
}
