package gov.nasa.gsfc.dqss;

import java.lang.reflect.Method;
import java.util.HashMap;

import ncsa.hdf.object.Dataset;

/*
 * MOD05_L2 NIR - the only one using for MOD05
 */
public class Modis1D2Q extends Mask {

	/* The procedure from going from hdfview QA NIR byte to Surface Type value:
	 * 1. Take positive of given number: -121 -> 121 = 01111001
	 * 2. Take two's complement:                       10000111
	 * 3. So since surface type is bits 7 and eight -121 translates to 2
	 *    (as does -123,-125 they vary because they have other conf values)
	 *    -61 -> 3 (00111101 -> 11000011)
	 *    +67 -> 1 (01000011)
	 * To obtain NIR TotPrecipConf which is bytes 2,3,4 you look at those bits after above procedure
	 * -125 -> 10000011 = 1
	 *  +67 -> 01000011 = 1
	 * -121 -> 10000111 = 3 
	 * To obtain Cloud Mask Cloudiness Flag bits 2,3
	 * 57 = 0 , 59 = 1, 61 = 2, 63 = 3
	 */    
	/*
	 * This class won't take care of both the IR and the NIR because the
	 * Cloud Mask is not needed for the IR. This is good because the Cloud_Mask_QA
	 * is 2030x1354 whereas IR is that divided by 5 or 406x270
	 * So that's how we can remember that CMQA is not needed for IR.
	 * Chris has modeled in his XML stub that the 
	 * IR QA is Total Precip Water Confidence Flag and the 
	 * MODIS Atmosphere QA Plan shows that the this info is in the first byte of the 
	 * Quality_Assurance_Infrared
	 * Chris has modeled that NIR will have quality variables 
	 * Quality_Assurance_Near_Infrared/Total Precip Water Confidence Flag 
	 * AND Cloud_Mask_QA/Cloud Mask Cloudiness flag. 
	 * The MODIS QA doc shows that the Total Precip Conf flag is in nits 2,3,4
	 * and the Cloudiness Flag is in bits 2,3. 
	 */
	
	 String[] direction = null;
	 DataFile file = null;
	public Modis1D2Q(Dataset data, Dataset[] qual, DataFile File , String dir) {
		super(data,qual,dir,File);
		file = File;
		if  (file.get_direction() != null) {
			direction = new String[file.get_direction().size()];
		}
		else {
			logger.error("No direction obtained for " + data.getName());
		}
	}

	 public void doMask_reflect() throws Exception
	  {
		 /*
		  * MOD05_L2 is so fast we can live with reflection
		  */
		  doMask();
	  }
	public void doMask() throws Exception {
		
    
    	if (quality != null ) {
		logger.debug("qa1:" + quality[0].getName());
		logger.debug("qa1:" + quality[1].getName());
    	}
    	else {
    		logger.error("No quality variables passed to Modis1D2Q");
    	}
    	byte[] qaData    = (byte[]) quality[0].read();
    	direction[0]   = file.get_direction().get(0); 
    	byte[] cldmskData = (byte[]) quality[1].read();
    	direction[1]   = file.get_direction().get(1); 
    	short fill = (short) MaskFactory.getFillValue().shortValue();
    	long[] start = data.getStartDims();
    	logger.debug("fill:" + fill);
    	logger.debug("rows * cols:" + rows * cols );
    	long[] pCrit = new long[file.crit2num.size()] ;
    	for (int i = 0; i < pCrit.length; ++i) {
    		pCrit[i] = file.crit2num.get(i);
    		logger.debug("Criteria levels sent:" + pCrit[i] + " " + direction[i]);
    	}
    	ReflectDirections rf = new ReflectDirections();
   	 
        Method qa = rf.getMethod(direction[0],Integer.TYPE, Long.TYPE);
        Object qalist[] = new Object[2];
        Method cm = rf.getMethod(direction[1],Integer.TYPE, Long.TYPE);
        Object cmlist[] = new Object[2];
        Object usefullist[] = new Object[2];
        
    	
    	qalist[1] = pCrit[0];
    	cmlist[1] = pCrit[1];
    	usefullist[1] = 0;
    	
    	

     	for (int layer = 0; layer < layers ; ++layer) {
    		if (start.length > 2) {
        		start[2] = layer;
        	}	
    		logger.debug("Masking Layer:" + layer);
    		
    	    short[] dataData = (short[]) data.read();
    	    logger.debug("length: data:" + dataData.length);
    	    stats.setpixels(rows, cols);
    	    
    	    for (int i = 0; i < rows * cols; ++i) {
    	    
    	    	// Use evaluateQualityFlag method when we have start and stop bits from ontology
    			qalist[0] = evaluateTotalPrecipConfFlag(qaData[i]);
    			cmlist[0] = evaluateCloudinessFlag(cldmskData[i]);
    			
    			/*
    			if (DQSS.delimiter.contains( "test" )) {
    				//testShiftingFirst(qaData[i],cldmskData[i]);
    				stats.countQAflags(evaluateTotalPrecipConfFlag(qaData[i]));
        			stats.countCldflags(evaluateCloudinessFlag(cldmskData[i]));
    			}
    			*/
    			//Suraiya said not to use usefulness
    			//usefullist[0] = evaluateUsefulnessBits(qaData[i]);
    			
    			if (dataData[i] == fill) {
    				dataData[i] = fill;
    	    		stats.countfill();
    			}
    			else if (
    					
    			/*(Boolean) qa.invoke(rf , usefullist)  && */
    	    	(Boolean) qa.invoke(rf , qalist)    	&&
    	    	(Boolean) cm.invoke(rf , cmlist)) { 
     	    		dataData[i] = 1;
    	    		stats.inc();
    	    		
    	    	}
    	    	else {
    	    		//logger.debug("filling");
    	    		//logger.debug(i + " qa:" + qalist[0] + " cldmsk" + cmlist[0]);
    	    		dataData[i] = fill;
    	    		
    	    	}
    			
    			
    		    
    		}
    		data.write(dataData);
	   }
    	stats.setTotal((int)(layers * rows * cols));
        //stats.publishQAFlags();		
        //stats.publishCldFlags();		
        
    		
    	
		
	}
	/* Java internally uses (or emulates) twos complement arithmetic. 
	 * Positive numbers are stored in binary. That means that for example the number 13 would be written 
	 * 0x0d = 00001101 = 1*8 + 1*4 + 0*2 + 1*1. 
	 * The high order bit is called the sign bit. For signed quantities such as byte, short, int and long, 
	 * when that high order bit is a 1, we consider the number negative. 
	 * To represent a negative number we first invert each bit, then add 1. 
	 * We would write -13 this way: 0x0d -  00001101 -> 11110010 -> 11110011 -> 0xf3.
	 * 
	 * two's complement of a number is obtained by taking negative (-num)
	 * 
	 * If number is negative then bit 7 (see MODAPS Atmosphere QA Plan for 005)
	 * is 1. To get the rest of the numbers take the negative and and it.
	 */
	
	/* The Water Vapor (or Total Precipitable Water) product combines results from
       both the NIR (1x1 km) and IR algorithms (5 x 5 km). For near infrared (NIR) total
       precipitable water, the first byte contains cloud mask QA (1 x 1 km), and the second
       byte (a separate array from the cloud mask QA) contains NIR product quality and
       retrieval processing flags. 
       For the five bytes of IR total precipitable water results, only
       product quality and retrieval processing flags are stored. Since IR total precipitable
       water results are copied from 07_L2, the cloud mask related QA flags can be retrieved
       from 07_L2, and therefore it will not be duplicated here in the 05_L2 product. 
    */
	/* Quality_Assurance_Near_Infrared: Length: 1 byte (8 bits)
	   Flag Name						Number of Bits		Bit Values		Definitions
	   Total Precipitable Water (NIR)	1					0,1				Not useful, Useful
			   Usefulness flag
		Total Precipitable Water (NIRT) 3					0,1,2,3,		Bad(Fill) Marginal,Good,VG
		Confidence Flag
		Inversion Method Used (NIR)     2                   0,1,2           2 chan ratio, 3 chan ratio, no retrival
		Surface Type                    2					0,1,2,3			Bright Land, Clear Ocean, Cloud, Sunglint
		
	*/
	private boolean evaluateQualityAssuranceNearInfraredBits(Byte a) {
		//byte[] bits = {1,2,4,8,16,32,64,-128};
		byte[] bits = {-128,64,32,16,8,4,2,1};
		byte usefulnessMask = 1;
		byte surfaceMask    = -64 ; // and the sign of the byte.
		byte confidenceMask = 2 + 4 + 8;
		byte test = 32 + 64;
		byte alwayszeroMask = 8;
		Integer b = a & 0xFF;
		
		
		int surface = a & surfaceMask;
		if (surface == 64) {
			surface = 1;
		}
		else if (surface == -128) {
			surface = 2;
		}
		else if (surface == -64) {
			surface = 3;
		}
		
		int usefulness = a & usefulnessMask;
		
		/* Confidence */
		int confidence = a & confidenceMask;
		if (confidence == 6) {
			confidence = 3;
		}
		else if (confidence == 2) {
			confidence = 1;
		}
		else if (confidence == 4) {
			confidence = 2;
		}
		else  if (confidence == 0) {
			
		}
		else {
			confidence = 9999;
		}
		if (confidence > 1 && usefulness == 1) {
			return true;
		}
		return false;
}

	private void evaluateBits(Byte a, int count) {
		//byte[] bits = {1,2,4,8,16,32,64,-128};
		byte[] bits = {-128,64,32,16,8,4,2,1};
		byte usefulnessMask = 1;
		byte surfaceMask    = -64 ; // and the sign of the byte.
		byte confidenceMask = 2 + 4 + 8;
		byte test = 32 + 64;
		byte alwayszeroMask = 8;
		Integer b = a & 0xFF;
		int line = count / 270;
		int pixel = count - 270 * line;
		int surface = a & surfaceMask;
		if (surface == 64) {
			surface = 1;
		}
		else if (surface == -128) {
			surface = 2;
		}
		else if (surface == -64) {
			surface = 3;
		}
		//System.out.print(a  + " ->  " + b + " = " );
		/* alwayszero test 4th bit is not used as confidence Flag is 0-3 but uses 3 bits*/
		int alwayszero = count & alwayszeroMask;
		if (alwayszero == 0) {
			//System.out.print (count + ",");
		}
		else {
			//System.out.print (" ,");
		}
		/* Usefulness: */
		int usefulness = a & usefulnessMask;
		//System.out.println ("Usefulness:" + usefulness);
		
		/* Confidence */
		int confidence = a & confidenceMask;
		if (confidence == 6) {
			confidence = 3;
		}
		else if (confidence == 2) {
			confidence = 1;
		}
		else if (confidence == 4) {
			confidence = 2;
		}
		else  if (confidence == 0) {
			
		}
		else {
			confidence = 9999;
		}
		System.out.println ("Confidence:" + confidence + " Usefulness:" + usefulness + 
				" surface:" + surface + " cnt:" + count + " :" + line + ":" + pixel);
		for (int i = 0; i < 8; ++i) {
			Integer value = bits[i] & confidenceMask; // this gives the twos complement
			//System.out.print(value  + ", ");
			
		}
		/* this gives the two's complement of the stored number which is already the two's complement
		 * of the actual number so it's what we want -- except -- signed bit... You get value of this
		 * bit by whether number is positive or negative.
		 */
		if (confidence == 9999 || surface == 3) {
		for (int i = 0; i < 8; ++i) {
			if (i == 0 ) {
				System.out.print(a + ":");
			}
			Integer value = bits[i] & a; // this gives the twos complement
			System.out.print(value  + ", ");
			if (i == 7 ) {
				System.out.println(" - " + count);		
			}
		}
		}
	}
	
}
