package gov.nasa.gsfc.dqss;

import gov.nasa.gsfc.Stopwatch;
import gov.nasa.gsfc.dqss.castor.Datavar;
import gov.nasa.gsfc.dqss.castor.QueryOutput;
import gov.nasa.gsfc.dqss.castor.QueryOutputChoiceSequence;
import gov.nasa.gsfc.dqss.castor.Root;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;

import ncsa.hdf.hdflib.HDFException;
import ncsa.hdf.object.Attribute;

import org.apache.log4j.Logger;
import org.exolab.castor.xml.MarshalException;
import org.exolab.castor.xml.Unmarshaller;
import org.exolab.castor.xml.ValidationException;

/*
 * This might be overkill, I could have gone straight to AirsMaskFactory, 
 * but let's see how it pans out.
 */
public class ProductFactory {
	
	DataFile file;
	String skip;
	MaskFactory MF;
	List<String> params;
	List<String> criteria;
	List<Statistics> stats4metadata;
	List<DQSSNumber> FillValues = null;
	String snvid;
	String inputFile;
	static Logger logger = Logger.getLogger(ProductFactory.class.getName());
	
	/*
	 * Constructor for GES-DISC operating under MODAPS Arch
	 * Ontology data file has not been created yet
	 */
	public ProductFactory(List<String> params,
			List<String> criteria, 
			List<Statistics> stats,
			
			String StagingDirectory,
			String ConfigDir,String S4PA_DatasetVersionID, String inputFile ) {
		if (DataFile.DetermineProduct(inputFile).equals("AIRS")) {
			 this.snvid = S4PA_DatasetVersionID;
			 this.stats4metadata = stats;
			 this.params = params;
			 this.criteria = criteria;
			 this.inputFile = inputFile;
			 
			 MF = new AirsMaskFactory(StagingDirectory,ConfigDir); 
			
			 // define logger first before doing this:
			 /*
			  *  do SDPreProcessing before opening file:
			  */
		     FillValues = SDPreProcessing(inputFile,params);
		     /*
		      * Open data file
		      */
			 file = new AIRSDataFile(inputFile,inputFile);
			 file.FileType = "AIRS";
			
		 }
		else if (DataFile.DetermineProduct(inputFile).equals("MODIS")) {
			 this.snvid = S4PA_DatasetVersionID;
			 this.stats4metadata = stats;
			 this.params = params;
			 this.criteria = criteria;
			 this.inputFile = inputFile;
			 
			 MF = new ModisMaskFactory(StagingDirectory,ConfigDir); 
			
			 // define logger first before doing this:
			 /*
			  *  do SDPreProcessing before opening file:
			  */
		     FillValues = SDPreProcessing(inputFile,params);
		     /*
		      * Open data file
		      */
			 file = new MODISDataFile(inputFile,inputFile);
			 file.FileType = "MODIS";
			
		 }
		 else {
			 logger.fatal("Unable to identify type of inputfile from:" + inputFile);
			 System.out.println("Unable to identify type of inputfile from:" + inputFile);
		 }
	}
	/*
	 * For MODAPS under MODAPS Arch:
	 * Ontology file has already been created.
	 */
	public ProductFactory(List<Statistics> stats,
			String StagingDirectory, String ConfigDir, 
			String inputFile ) {
		
		params = new ArrayList<String>();
		criteria = new ArrayList<String>();
		
		if (DataFile.DetermineProduct(inputFile).equals("AIRS")) {

			
			
			 this.stats4metadata = stats;
			 this.inputFile = inputFile;
			 
			 MF = new AirsMaskFactory(StagingDirectory,ConfigDir); 
			
			 // define logger first before doing this:
			
			 
			 
		     /*
		      * Open data file
		      */
			 file = new AIRSDataFile(inputFile,inputFile);
			 file.FileType = "AIRS";
			
		 }
		else if (DataFile.DetermineProduct(inputFile).equals("MODIS")) {

			
			
			 this.stats4metadata = stats;
			 this.inputFile = inputFile;
			 
			 MF = new ModisMaskFactory(StagingDirectory,ConfigDir); 
			
			 // define logger first before doing this:
			
			 
			 
		     /*
		      * Open data file
		      */
			 file = new MODISDataFile(inputFile,inputFile);
			 file.FileType = "MODIS";
			
		 }
		 else {
			 logger.fatal("Unable to identify type of inputfile from:" + inputFile);
			 System.out.println("Unable to identify type of inputfile from:" + inputFile);
		 }
	}
	/*
	 * Constructor for MODAPS operating under MODAPS Arch
	 * used by OntologyGetter
	 */
	public ProductFactory(List<String> params,List<String> criteria, 	
			String ConfigDir,String stagingdir, String S4PA_DatasetVersionID,
			String inputfile) {
			 this.snvid = S4PA_DatasetVersionID;
			 this.params = params;
			 this.criteria = criteria;
			 this.inputFile = inputfile;
			 
			 if (DataFile.DetermineProduct(inputFile).equals("AIRS")) {
				 MF = new AirsMaskFactory(stagingdir, ConfigDir);
				
			 }
			 else if (DataFile.DetermineProduct(inputFile).equals("MODIS")) {
				 MF = new ModisMaskFactory(stagingdir, ConfigDir);
				
			 }
			 else {
				 logger.fatal("Unable to identify type of inputfile from:" + inputfile);
				 System.err.println("Unable to identify type of inputfile from:" + inputfile);
			 }
			 logger.debug("Filetype:" + DataFile.DetermineProduct(inputFile));
	}
		 
	public void getParamsForPreProcessing() {
		
	
		FileReader reader = null;
	    try {
	    	
	        reader = new FileReader(DQSS.OntFile);
	        
	    } catch (FileNotFoundException e) {
	        // TODO Auto-generated catch block
	        e.printStackTrace();
	        logger.error("ontology file not found:" + DQSS.OntFile);
	        System.exit(2);
	    } catch (NullPointerException e) {
	    	logger.warn("Ontology file was null");
	    	System.exit(2);
	    }
	     
	   
	    Root xmlstuff = null;
		try {
		    xmlstuff =
		    (Root) Unmarshaller.unmarshal(Root.class, reader);
		} catch (MarshalException e) {
		    // TODO Auto-generated catch block
			String explan = "Data in ontology file not quite right. Check arguments sent to ontology";
			logger.error(explan);
			System.err.println(explan);
		    e.printStackTrace();
		    System.err.println(explan);
		    System.exit(3);
		} catch (ValidationException e) {
		    // TODO Auto-generated catch block
		    e.printStackTrace();
		}
		try {
			xmlstuff.validate();
		} catch (ValidationException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		
        Datavar[] dv = xmlstuff.getDatavar();
		
		Enumeration<? extends Datavar> items = xmlstuff.enumerateDatavar();
		for (; items.hasMoreElements()  ;)  {
			
			
		    Datavar it = items.nextElement();
		    
		    QueryOutput[] qos = it.getQueryOutput();
		    String view = qos[0].getQualityView();
		    logger.debug("View:" + view);
		    if (view != null) {
		    	params.add(view.replace("View", ""));
		    }
		    else {
		    	logger.fatal("Could not retrieve  parameter from XML ontology file");
		    }
		}
	}
	

	public void runWithModisOntologyFile() {
		
		MODISDataFile afile = (MODISDataFile) file;
		FileReader reader = null;
		
		/*
		  *  do SDPreProcessing before opening file:
		  *  This is done already when ges-disc arch is used.
		  */
		
		 if (params == null || params.size() == 0) {
			 getParamsForPreProcessing();
		 }
		 if (FillValues == null || FillValues.size() == 0 ) {
			 FillValues = SDPreProcessing(inputFile,params);
		}
		 //TransferMissingAttributes(inputFile); // in this location only for running under debug 
	    try {
	        //reader = new FileReader("/home/rstrub/temp2.xml");
	        reader = new FileReader(DQSS.OntFile);
	        logger.debug("Using ontology file:" + DQSS.OntFile);
	    } catch (FileNotFoundException e) {
	        // TODO Auto-generated catch block
	        e.printStackTrace();
	        logger.error("ontology file not found:" + DQSS.OntFile);
	    }
	   
	    Root xmlstuff = null;
		try {
		    xmlstuff =
		    (Root) Unmarshaller.unmarshal(Root.class, reader);
		} catch (MarshalException e) {
		    // TODO Auto-generated catch block
		    e.printStackTrace();
		} catch (ValidationException e) {
		    // TODO Auto-generated catch block
		    e.printStackTrace();
		}
		try {
			xmlstuff.validate();
		} catch (ValidationException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		logger.debug("FillValues Size:" + FillValues.size());
		Iterator fillitr = FillValues.iterator();
	   		
        Datavar[] dv = xmlstuff.getDatavar();
		
		Enumeration<? extends Datavar> items = xmlstuff.enumerateDatavar();
		for (; items.hasMoreElements()  ;)  {
			
			DQSSNumber fillvalue = (DQSSNumber) fillitr.next();
		    Datavar it = items.nextElement();
		    
		    QueryOutput[] qos = it.getQueryOutput();
		   
		    Enumeration<? extends QueryOutput> qvs = it.enumerateQueryOutput();
		    
		    int nQSVars = it.getQueryOutputCount();
		    
		    
		    afile.nCoordCons=0;
		    afile.nQualLevel=0;
		    afile.nConsRef=0;
		    afile.nRef=0;
		    
		    afile.qs_names.clear();
		    afile.ref_names.clear();
		    afile.qs_types.clear();
		    afile.qs_reftypes.clear();
		    afile.mins.clear();
		    afile.maxs.clear();
		    afile.crit2num.clear();
		   
		    for (int i = 0; i < nQSVars; ++i) {
		    	afile.updateMember(file.qs_names,"ScreeningVariable Name",  qos[i].getScreeningVariable());
		    	afile.updateMember(file.ref_names,"ReferenceVariable Name",  qos[i].getQueryOutputChoice().getReferenceVariable());
		    	afile.updateMember(file.qs_types,"ScreeningVariable Type",     qos[i].getScreenVarType());
		    	afile.updateMember(file.qs_reftypes,"ReferenceVariable Type",  qos[i].getRefVarType());
		    	afile.updateLongMember(file.crit2num,"User's Criteria", qos[i].getLevel());
		    	//QueryOutputSequence qs = qos[i].getQueryOutputSequence();
		    	afile.updateMember(file.qs_coordcon_dir,"coordinate_constraint relationship",
		    			qos[i].getConstraint_relationship());
		    	afile.updateMember(file.qs_direction,"Direction",             qos[i].getDirection());
		    	
		    	QueryOutputChoiceSequence qs = qos[i].getQueryOutputChoice().getQueryOutputChoiceSequence();
				    if (qs != null) {
				    	afile.updateMember(file.dims,"Dimension", qs.getDim());
				    	afile.updateLongMember(file.mins, "Minimum DataSlice", qs.getMin());
				    	afile.updateLongMember(file.maxs, "Max DataSlice", qs.getMax());
				    	
				    }
		            logger.debug(":" + afile.nConsRef + ":" + afile.nRef + ":" + afile.nCoordCons + ":" + afile.nQualLevel);
		        if (afile.qs_reftypes.size() > 0) {
			    	afile.qs_names.add(afile.ref_names.get(0));
			    	afile.qs_types.add(afile.qs_reftypes.get(0));
			    }
		     
		    }
		    // Viewname will be same in all QueryOutput[]
		    String dataVariable = qos[0].getQualityView();
		    dataVariable = dataVariable.replace("View", "");
		    logger.warn("Processing DataVariable:" + dataVariable);
		   // DQSSNumber fillvalue = getFillValueN(inputFile,dataVariable);
		    // this has already been done! params.add(dataVariable);
		    //String UsersOriginalCriteria = afile.ReEngCriteria(qos);
		    String UsersOriginalCriteria = afile.ReEngCriteria(qos);
		    criteria.add(UsersOriginalCriteria);
		    logger.debug("criteria array problem size:" + criteria.size());
		    logger.debug("criteria array problem:" + UsersOriginalCriteria );//criteria.add(UsersOriginalCriteria);
		    //logger.debug("criteria array problem size:" + criteria.size());
		    //logger.debug("criteria array problem:" +UsersOriginalCriteria );
		    //FillValues.add(fillvalue);
		    MF.dataVariable = dataVariable;
		    Statistics stats = MF.StartMasking(afile, fillvalue);
		    if (stats != null) {
	    		   stats4metadata.add(stats);
	    	}
		   
		}
		WriteStatisticsWithDataFile(stats4metadata,file);
		/*
		 * All H4File/Dataset library stuff has been run. It needs to be closed
		 * so that HDFLibrary stuff can reopen for write (it opens it in a different way)
		 */
		file.close();  
		
	}
	
	public void runWithOntologyFile() {
		if (file.FileType == "AIRS") {
			runWithAirsOntologyFile();
		}
		else if (file.FileType == "MODIS") {
			runWithModisOntologyFile();
		}
	}
	public void runWithAirsOntologyFile() {
		
		AIRSDataFile afile = (AIRSDataFile) file;
		FileReader reader = null;
		
		/*
		  *  do SDPreProcessing before opening file:
		  *  This is done already when ges-disc arch is used.
		  */
		 if (params == null || params.size() == 0) {
			 getParamsForPreProcessing();
		 }
		 if (FillValues == null || FillValues.size() == 0 ) {
			 FillValues = SDPreProcessing(inputFile,params);
		}
		 if (false ) {
			 file.close(); // we want to open the Vdata with read access.
			 VDataManagement VD = new VDataManagement(inputFile ,"r"); // in this location only for running under debug
			 String classname = VD.ReadSomeVData("_FV_olr");
			 VD.close();
			 //VD = new VDataManagement(inputFile ,"r");
			 //VD.CreateVData("_VF_olr_orig",classname);
			 //VD.close();
		 
		 
			try {
				SDDataFile sid = new SDDataFile(inputFile);
				 sid.CreateVData("_VF_olr_orig",classname);
				 sid.destroy();
			} catch (HDFException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}
		 }
	    try {
	        //reader = new FileReader("/home/rstrub/temp2.xml");
	        reader = new FileReader(DQSS.OntFile);
	        logger.debug("Using ontology file:" + DQSS.OntFile);
	    } catch (FileNotFoundException e) {
	        // TODO Auto-generated catch block
	        e.printStackTrace();
	        logger.error("ontology file not found:" + DQSS.OntFile);
	    }
	   
	    Root xmlstuff = null;
		try {
		    xmlstuff =
		    (Root) Unmarshaller.unmarshal(Root.class, reader);
		} catch (MarshalException e) {
		    // TODO Auto-generated catch block
		    e.printStackTrace();
		} catch (ValidationException e) {
		    // TODO Auto-generated catch block
		    e.printStackTrace();
		}
		try {
			xmlstuff.validate();
		} catch (ValidationException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		Iterator fillitr = FillValues.iterator();
   		
       
		
        Datavar[] dv = xmlstuff.getDatavar();
		
		Enumeration<? extends Datavar> items = xmlstuff.enumerateDatavar();
		for (; items.hasMoreElements()  ;)  {
			
			DQSSNumber fillvalue = (DQSSNumber) fillitr.next();
			
		    Datavar it = items.nextElement();
		    
		    QueryOutput[] qos = it.getQueryOutput();
		   
		    Enumeration<? extends QueryOutput> qvs = it.enumerateQueryOutput();
		    
		    int nQSVars = it.getQueryOutputCount();
		    
		    
		    afile.nCoordCons=0;
		    afile.nQualLevel=0;
		    afile.nConsRef=0;
		    afile.nRef=0;
		    
		    afile.qs_names.clear();
		    afile.ref_names.clear();
		    afile.qs_types.clear();
		    afile.qs_reftypes.clear();
		    afile.mins.clear();
		    afile.maxs.clear();
		    afile.crit2num.clear();
		    afile.qs_direction.clear();
		    afile.qs_coordcon_dir.clear();
		    
		    for (int i = 0; i < nQSVars; ++i) {
		    	afile.updateMember(file.qs_names,"ScreeningVariable Name",    qos[i].getScreeningVariable());
		    	//afile.updateMember(file.ref_names,"ReferenceVariable Name",   qos[i].getReferenceVariable());
		    	afile.updateMember(file.ref_names,"ReferenceVariable Name",   qos[i].getQueryOutputChoice().getReferenceVariable());
		    	afile.updateMember(file.qs_types,"ScreeningVariable Type",    qos[i].getScreenVarType());
		    	
		    		afile.updateMember(file.qs_reftypes,"ReferenceVariable Type", qos[i].getRefVarType());
		    		afile.updateMember(file.qs_coordcon_dir,"coordinate_constraint relationship",
			    			qos[i].getConstraint_relationship());
		    	
		    	afile.updateMember(file.qs_direction,"Direction",             qos[i].getDirection());
		    	
		    	afile.updateLongMember(file.crit2num,"User's Criteria", qos[i].getLevel());
		    	
	    	QueryOutputChoiceSequence qs = qos[i].getQueryOutputChoice().getQueryOutputChoiceSequence();
				    if (qs != null) {
				    	afile.updateMember(file.dims,"Dimension", qs.getDim());
				    	afile.updateLongMember(file.mins, "Minimum DataSlice", qs.getMin());
				    	afile.updateLongMember(file.maxs, "Max DataSlice", qs.getMax());
				    	
				    }
		            logger.debug(":" + afile.nConsRef + ":" + afile.nRef + ":" + afile.nCoordCons + ":" + afile.nQualLevel);
		        if (afile.qs_reftypes.size() > 0) {
			    	afile.qs_names.add(afile.ref_names.get(0));
			    	afile.qs_types.add(afile.qs_reftypes.get(0));
			    }
		    }
		    // Viewname will be same in all QueryOutput[]
		    String dataVariable = qos[0].getQualityView();
		    dataVariable = dataVariable.replace("View", "");
		    logger.warn("Processing DataVariable:" + dataVariable);
		    //float fillvalue = getFillValue(inputFile,dataVariable);
		    // this has already been done! params.add(dataVariable);
		    String UsersOriginalCriteria = afile.ReEngCriteria(qos);
		    criteria.add(UsersOriginalCriteria);
		    logger.debug("criteria array problem:" + criteria.size());
		    logger.debug("criteria array problem:" +UsersOriginalCriteria );
		    
		    /*
		    try {
				reflect(file.qs_direction);
			} catch (SecurityException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (IllegalArgumentException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (NoSuchMethodException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (Throwable e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		    */
		    MF.dataVariable = dataVariable;
		    Statistics stats = MF.StartMasking(afile, fillvalue);
		    if (stats != null) {
	    		   stats4metadata.add(stats);
	    	}
		}
		WriteStatisticsWithDataFile(stats4metadata,file);
		/*
		 * All H4File/Dataset library stuff has been run. It needs to be closed
		 * so that HDFLibrary stuff can reopen for write (it opens it in a different way)
		 */
		file.close();  
		
	}
	
	
	public void run() {
		  
		
		   /*
			 * Mask each data variable:
			 */
		    Stopwatch s = new Stopwatch();
		    s.start();
		   	Iterator itr = params.iterator();
		   	Iterator crititr = criteria.iterator();
		   	Iterator fillitr = FillValues.iterator();
		   	while(itr.hasNext()) {
		   		
		    	String param = (String) itr.next();
		    	String criterion = (String) crititr.next(); 
		    	if ( ! criterion.equals(skip)) {
		    	DQSSNumber fillvalue = (DQSSNumber) fillitr.next();
			    	if (paramtest(param, criterion)) {
			    	   MF.PrepMasking(file, snvid,  param, criterion, 
				    			   fillvalue);
			    		
			    	   Statistics stats = MF.StartMasking(file, fillvalue);
			    	   if (stats != null) {
			    		   stats4metadata.add(stats);
			    	   }
			    	}
		    	}
		    }
		   	s.stop();
		   	logger.debug("Total Masking and ontology: elapsed time: " + s.getElapsedTime());
		   	s.reset();
		   	s.start();
		   	WriteStatisticsWithDataFile(stats4metadata,file);
		   	s.stop();
		   	
		   	logger.debug("Write Stats: elapsed time: " + s.getElapsedTime());
		    file.close();
		    
	}
	
	/*
	 * Get the ontology data for all of the datavariables to be screened.
	 */
	public void LoadOntologyData() {
	
		   	Iterator itr = params.iterator();
		   	Iterator crititr = criteria.iterator();
		   	
		   	String xml = "";
		   	while(itr.hasNext()) {
		   		
		    	String param = (String) itr.next();
		    	String criterion = (String) crititr.next(); 
		    	
		    	    if (snvid == null) {
		    	    	logger.fatal("snvid is null");
		    	    	return;
		    	    }
		    	    if (param == null) {
		    	    	logger.fatal("datavariable is null for:" + snvid);
		    	    	return;
		    	    }
		    	    if (criterion == null) {
		    	    	logger.fatal("usercritia  is null for:" + snvid + ", " + param);
		    	    	return;
		    	    }
		    	    
			    	if (paramtest(param, criterion)) {
			    		logger.debug("Sending:" + snvid + "," + param + "," + criterion + " for ontology loading");
			    	    xml += MF.LoadOntology(snvid,  param, criterion);
			    	    xml += "\n";
			    	}
		    	
		    }
		    xml = "<root>\n" + xml + "\n</root>";
		   
		    /*
		     * DQSS.OntFile will just be a basename unless it is given, in which
		     * case it will already exist and this code won't be called.
		     */
		    DQSS.fileOverWrite(DQSS.OntFile, xml);
	}
	
	/*
	 * MODAPS Arch:
	 */
	public DQSSNumber getFillValue(String inputFile, String datasetName) 
	{
		DQSSNumber fill = new DQSSNumber(-66664.2);
		try {
			SDDataFile sid = new SDDataFile(inputFile,"READONLY");
			
		    fill= sid.getFillValue(datasetName, "_FillValue");
			
		   	sid.destroy();
		} catch (HDFException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		
		return fill;
	}
	public DQSSNumber getFillValueN(String inputFile, String datasetName) 
	{
		DQSSNumber fill = new DQSSNumber(-7774.2);
	    
		try {
			SDDataFile sid = new SDDataFile(inputFile,"READONLY");
			
		    fill= sid.getFillValue(datasetName, "_FillValue");
			
		   	sid.destroy();
		} catch (HDFException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		return fill;
	}
	/*public List<Float> AirsSDPreProcessing(String inputFile, List<String> params) 
	{
		List<Float> fillvalues = new ArrayList<Float>();
		
		try {
			SDDataFile sid = new SDDataFile(inputFile); 
			Iterator itr = params.iterator();
		   	while(itr.hasNext()) {
		    	String datasetName = (String) itr.next();  
		    	fillvalues.add(sid.getFillValue(datasetName, "_FillValue"));
			}
		   	sid.destroy();
		} catch (HDFException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		
		return fillvalues;
	}
	*/
	public List<DQSSNumber> SDPreProcessing(String inputFile, List<String> params) 
	{
		List<DQSSNumber> fillvalues = new ArrayList<DQSSNumber>();
		if (params == null || params.size() < 1) {
			logger.error("Parameter list sent to look for fillvalues is empty");
		}
		try {
			SDDataFile sid = new SDDataFile(inputFile); 
			Iterator itr = params.iterator();
			int i=0;
		   	while(itr.hasNext()) {
		    	String datasetName = (String) itr.next();  
		    	fillvalues.add(sid.getFillValue(datasetName, "_FillValue"));
		    	logger.debug("Added fillvalue:" + fillvalues.get(i) + " for " + datasetName);
		    	++i;
			}
		   	sid.destroy();
		} catch (HDFException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		
		return fillvalues;
	}
	public boolean paramtest(String param, String criterion) {
		if (param.length() < 1) {
			logger.error("datavariable to screen has length 0");
			return false;
		}
		if (criterion.length() < 1) {
			logger.error("criterion has length 0");
			return false;
		}
		if (criterion.equals(DQSS.NoScreeningString)) {
			logger.debug("Not Screening " + param);
			return false;
		}
		logger.debug("Screening datavariable:" + param + " with " + criterion);
		return true;
	}
	
	/*
	 * Using ncsa.hdf.object.FileFormat for this operation works!
	 */
	void WriteStatisticsWithDataFile (List<Statistics> stats4metadata, DataFile file)  {
		String ScreeningStats = "\n";
		String ScreeningAttribute = "Quality Screening";
		 Iterator statsitr = stats4metadata.iterator();
		
		 while(statsitr.hasNext()) {
			 Statistics nextstats = (Statistics) statsitr.next();
			 int qlv = 0;
			 for (int i = 0; i < nextstats.qualityVariables.size(); ++i) {
				 
				 ScreeningStats += nextstats.getVariable() + " was filtered using " + 
			     nextstats.getQualityVariable(i).trim() ;
				 if (nextstats.getQualityVariableType(i).contains("QualityLevelVariable")) {
					 if (nextstats.getQualityLevel(qlv) != null) {
						 logger.debug("QualityLevel:" + nextstats.getQualityLevel(qlv));
						 ScreeningStats += " based on " + nextstats.getQualityLevel(qlv).trim() + ". " ;
						 ++qlv;
					 }
				 }
				 else {
					 ScreeningStats += ". ";
				 }
			 }
			 	 ScreeningStats += "  Non-fill data before screening: " + nextstats.getPrePercentage() + "%;" + 
			     " after screening " + nextstats.getPercentage()  + "%.\n";
				 logger.debug(ScreeningStats);
			 
		 }
		 Attribute stats_attr;
		try {
			
			stats_attr = file.getYourAttribute(ScreeningAttribute,ScreeningStats.length() );
			logger.debug("Screening Stats length:" + ScreeningStats.length());
			String[] StatsArray = {ScreeningStats};
			file.writeStatsToMetadata(stats_attr,StatsArray);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		 
				
	}
	/*
	 * MODAPS Arch:
	 */
	void SDPostProcessing(String inputFile)
	    {
		
		if (criteria == null) {
			logger.warn("No criteria array available");
			
		}
		else {
			logger.debug("criteria problem: size 0? : " + criteria.size());
			logger.debug("criteria problem: empty ? : " + criteria.get(0));
			logger.debug("params? : " + params.get(0));
		}
		if (params == null) {
			logger.error("No params array available");
			return;
		}
		if (FillValues == null) {
			logger.warn("No FillValues array available");
			return;
		}
		
	    try {
		    SDDataFile sid = new SDDataFile(inputFile);
		    Iterator<String> itr = params.iterator();
		    Iterator<String> crititr = criteria.iterator();
		    Iterator<DQSSNumber> fvitr   = FillValues.iterator();
		    if (criteria.size() == 0) {
		    	logger.error("Lost criteria");
		    }
		    int i = 0;
		    while(itr.hasNext()) {
		    	String name = (String) itr.next();
		    	/*
		    	 * fill in dimension names for newly created masks
		    	 */
		    	String criterion = null;
		    	String check[] = null;
		    	if (name.length() > 0) {
			    	try {
				    	String[] newdimnames = sid.GetDimensionNames(name);
				    	sid.UpdateDimensionNames(name + DQSS.MaskName, newdimnames);
				    	/*
				    	 * Set the quality level attribute
				    	 */
				    	criterion = (String) crititr.next();
				    	DQSSNumber fillvalue = (DQSSNumber) fvitr.next(); 
				    	logger.debug("crit:" + criterion);
			    		
			    		
				    	sid.SetAttribute(name + DQSS.MaskName, "Quality", criterion);
				    	check = sid.GetDimensionNames(name + DQSS.MaskName);
				    	
				    	logger.debug("fill:" + fillvalue);
				    	sid.SetFill(name + DQSS.MaskName, fillvalue);
			    	} catch (NoSuchElementException e) {
			    		
			    		logger.error("Elements for SDPostProcessing name, criterion, dimname (" + i + ")");
			    		logger.error(name);
			    		
			    		
			    	}
		    	}
		    	++i;
		    	
		    	
		    }
		    
		    sid.destroy();
	    }
	    catch (HDFException e) {
	    	e.printStackTrace();
	    	logger.error("SD stuff failed");
	    }
	    
	}
	
	void SetSpecificAttributesInMask(String inputFile)
    {
	/*
	 * Attributes I want to set in mask:
	 * Valid Range: 1 and -9999
	 * long_name    : prepend "mask for"
	 * ms
	 * add_offset = 0
	 * Scale factor should be 1.
	 */

	if (params == null) {
		logger.error("No params array available. ie no dataset names with which to retrieve ids");
		return;
	}
	
    try {
	    SDDataFile sid = new SDDataFile(inputFile);
	    Iterator<String> itr = params.iterator();
	  
	    int i = 0;
	    while(itr.hasNext()) {
	    	String name = (String) itr.next();
	    	/*
	    	 * fill in dimension names for newly created masks
	    	 */
           // name = "_FV_" + name;
	    	if (name.length() > 0) {
		    	try {

		    		
				    HashMap<String,Integer[]> nameAndLen = new HashMap<String,Integer[]>();
				    sid.GetListOfAttributeNames(name, nameAndLen);
				    Iterator it = nameAndLen.entrySet().iterator();
				   
				    while (it.hasNext()) {
				       try {
				    	Map.Entry pairs = (Map.Entry)it.next();
				    	String attr_name =    (String)pairs.getKey();
				    	String attr_value = "";
				    	
				        Object attributevalue = sid.GetAttribute(name, attr_name, (Integer[])pairs.getValue());
				        Integer[]   dimensions = (Integer[])pairs.getValue();
				        
				        
				        if (attr_name.matches("long_name|Cell.*|valid.*")) 
				        {
				        	if (attr_name.contains("long_name")) {
				        		String prepended  = "mask for " + (String)attributevalue;
				        		attributevalue = new Object();
				        		attributevalue = prepended;
				        	}
				        	
				        	
					        if (dimensions[0] == 4) {
					        logger.debug ("Attribute " + attr_name + " has value:" + attr_value);
					        }
					        else if  (dimensions[0] == 22) {
					        	Short[] IntVals = (Short[])attributevalue;
					        	if (attr_name.contains("valid")) {
					        		IntVals[0] = FillValues.get(0).shortValue();
					        		IntVals[1] = 1;
					        		attributevalue = new Object();
					        		attributevalue = IntVals;
					        	}
						        logger.debug ("Attribute " + attr_name + " last value is:" + IntVals[dimensions[1]-1]);
						        }
					        else if  (dimensions[0] == 24) {
					        	Integer[] IntVals = (Integer[])attributevalue;
						        logger.debug ("Attribute " + attr_name + " last value is:" + IntVals[dimensions[1]-1]);
						        }
					        else if  (dimensions[0] == 6) {
					        	Double[] IntVals = (Double[])attributevalue;
						        logger.debug ("Attribute " + attr_name + " last value is:" + IntVals[dimensions[1]-1]);
						        }
					        else if  (dimensions[0] == 5) {
					        	Float[] IntVals = (Float[])attributevalue;
						        logger.debug ("Attribute " + attr_name + " last value is:" + IntVals[dimensions[1]-1]);
						        }
					        else {
					        	logger.error("Need code for type: " + dimensions[0]);
					        }
					        sid.SetAllAttributes(name + DQSS.MaskName, attr_name ,dimensions,attributevalue);
					        // done in Screener since *_orig don't exist yet: sid.SetAllAttributes(name + DQSS.OrigName, (String)pairs.getKey() ,dimensions,attributevalue);
				        }
				       }
				       catch (HDFException e) {
				    	   logger.error("Datatype problem with attributes");
				       }
				    }
			    	
			    	
			    	
		    	} catch (NoSuchElementException e) {
		    		
		    		logger.error("Elements for SDPostProcessing name, criterion, dimname (" + i + ")");
		    		logger.error(name);
		    		
		    		
		    	}
	    	}
	    	++i;
	    	
	    	
	    }
	    
	    sid.destroy();
    }
    catch (HDFException e) {
    	e.printStackTrace();
    	logger.error("SD stuff failed");
    }
    
}
	void TransferMissingAttributes(String inputFile)
    {
	

	if (params == null) {
		logger.error("No params array available. ie no dataset names with which to retrieve ids");
		return;
	}
	
    try {
	    SDDataFile sid = new SDDataFile(inputFile);
	    Iterator<String> itr = params.iterator();
	  
	    int i = 0;
	    while(itr.hasNext()) {
	    	String name = (String) itr.next();
	    	/*
	    	 * fill in dimension names for newly created masks
	    	 */
           // name = "_FV_" + name;
	    	if (name.length() > 0) {
		    	try {
		    		/*
					 * This returns the attribute name, datatype and string length
					 */
				   
				    HashMap<String,Integer[]> nameAndLen = new HashMap<String,Integer[]>();
				    sid.GetListOfAttributeNames(name, nameAndLen);
				    Iterator it = nameAndLen.entrySet().iterator();
				   
				    while (it.hasNext()) {
				       try {
				    	Map.Entry pairs = (Map.Entry)it.next();
				        Object attributevalue = sid.GetAttribute(name, (String)pairs.getKey(), (Integer[])pairs.getValue());
				        Integer[]   dimensions = (Integer[])pairs.getValue();
				        if (dimensions[0] == 4) {
				        logger.debug ("Attribute " + (String)pairs.getKey() + " has value:" + (String)attributevalue);
				        }
				        else if  (dimensions[0] == 22) {
				        	Short[] IntVals = (Short[])attributevalue;
					        logger.debug ("Attribute " + (String)pairs.getKey() + " last value is:" + IntVals[dimensions[1]-1]);
					        }
				        else if  (dimensions[0] == 24) {
				        	Integer[] IntVals = (Integer[])attributevalue;
					        logger.debug ("Attribute " + (String)pairs.getKey() + " last value is:" + IntVals[dimensions[1]-1]);
					        }
				        else if  (dimensions[0] == 6) {
				        	Double[] IntVals = (Double[])attributevalue;
					        logger.debug ("Attribute " + (String)pairs.getKey() + " last value is:" + IntVals[dimensions[1]-1]);
					        }
				        else if  (dimensions[0] == 5) {
				        	Float[] IntVals = (Float[])attributevalue;
					        logger.debug ("Attribute " + (String)pairs.getKey() + " last value is:" + IntVals[dimensions[1]-1]);
					        }
				        else {
				        	logger.error("Need code for type: " + dimensions[0]);
				        }
				        sid.SetAllAttributes(name + DQSS.MaskName, (String)pairs.getKey() ,dimensions,attributevalue);
				        // done in Screener since *_orig don't exist yet: sid.SetAllAttributes(name + DQSS.OrigName, (String)pairs.getKey() ,dimensions,attributevalue);
				       }
				       catch (HDFException e) {
				    	   logger.error("Datatype problem with attributes");
				       }
				    }
			    	
			    	
			    	
		    	} catch (NoSuchElementException e) {
		    		
		    		logger.error("Elements for SDPostProcessing name, criterion, dimname (" + i + ")");
		    		logger.error(name);
		    		
		    		
		    	}
	    	}
	    	++i;
	    	
	    	
	    }
	    
	    sid.destroy();
    }
    catch (HDFException e) {
    	e.printStackTrace();
    	logger.error("SD stuff failed");
    }
    
}
	void reflect(List<String> dirs) throws SecurityException, NoSuchMethodException, IllegalArgumentException, Exception, Throwable {
		
		Stopwatch s = new Stopwatch();

	    s.start();
	 	int temp2 = 0;
	 	for (int pixel = 0 ; pixel < 250000; ++pixel ) {
	          
	 		 if (pixel < 250000) {
	 				 temp2 = pixel;
	 					  
	 		 }
	 			   
	 	}
	    System.out.println("Elapsed Time in millisecs: " + s.getElapsedTime());
	           
		ReflectDirections rf = new ReflectDirections();
		Class c = null;
		try {
			c = Class.forName("gov.nasa.gsfc.dqss.ReflectDirections");
		} catch (ClassNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		 Method m[] = c.getDeclaredMethods();
		 Class partypes[] = new Class[2];
         partypes[0] = Integer.TYPE;
         partypes[1] = Integer.TYPE;
           
		 Method meth = c.getMethod("lt", partypes);
		 
		   s.reset();
		 s.start();
		 int temp = 0;
		 boolean retval;
		 Object arglist[] = new Object[2];
		 for (int pixel = 0 ; pixel < 250000; ++pixel ) {
         	 
        	 arglist[0] = pixel;
        	 arglist[1] = 9980;
			 Object resp = meth.invoke(rf,arglist);
		     retval = (Boolean)resp;
			 if (retval) {
				 temp = pixel;
					   //System.out.println(pixel);
			 }
			   
		  }
          System.out.println("Elapsed Time in millisecs: " + s.getElapsedTime());
		
       
	}
	
	
	
}
