package gov.nasa.gsfc.dqss.ontology_stub;

import gov.nasa.gsfc.dqss.Masker;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import org.dom4j.Document;
import org.dom4j.DocumentHelper;
import org.dom4j.Element;
import org.dom4j.XPath;

import com.hp.hpl.jena.ontology.OntModel;
import com.hp.hpl.jena.ontology.OntModelSpec;
import com.hp.hpl.jena.query.Query;
import com.hp.hpl.jena.query.QueryExecution;
import com.hp.hpl.jena.query.QueryExecutionFactory;
import com.hp.hpl.jena.query.QueryFactory;
import com.hp.hpl.jena.query.QuerySolution;
import com.hp.hpl.jena.query.ResultSet;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.RDFNode;

/*
 * Algorithm: inputs: parameter, constraints
 *            outputs: query results xml
 *            1. findQuery - do canned SPARQL query do determine the query that will be done.
 *             .             populate this.Key.
 *            2. getquery  - populate this.Query using this.Key
 *            2. fillQuery - using the key interpolate the this.Query
 *            3. doQuery   - run the query and populate this.Results
 *            4. intoXML   - format results;
 */
public class Ontology {
	
	String ontLocation = "http://mirador-ts2.gsfc.nasa.gov/DataQualityViews.owl";
	OntModel dqss;
	String Query;
	String Key;
	Document Results;
	int Status = 0;
	
	public  Ontology(String dataset, String dataVariable, String criteria ){
		
		Results = DocumentHelper.createDocument();
		stubVenku(dataset, dataVariable, criteria);
		
		 // guess if we're using a daml source
	}
	
	private void stubVenku(String dataset, String dataVariable, String criteria)
	{
		
        
        URL xmlbase = null;
        try {
        	xmlbase = new URL(ontLocation);
        }
        catch (MalformedURLException e){}
        
        /*
        
         * Create ontology model for parameter ontology
         */
         dqss = ModelFactory.createOntologyModel(OntModelSpec.OWL_MEM, null);
             
         Masker.ReportError("ns:" + dqss.getNsPrefixMap(),2);
        /* 
         * Create ontology model for mirador - datasets ontology
         */
         dqss.setNsPrefix( "dqssv", "http://org.esipfed.dataquality/dqss#");
        
        dqss.read(ontLocation);
        getQuery(dataset, dataVariable, criteria);
    	fillQuery(dataVariable,criteria);
    	doQuery();
        
	}
	
	private void setStatus(int status) {
		this.Status = status;
	}
    
    
      
    
     void getQuery(String dataset, String dataVariable, String criteria) {
    	 
    	 if (dataVariable.startsWith("TAirMWOnly")) {
    		 this.Query = "PREFIX dqssv: <http://org.esipfed.dataquality/dqss#> " +
				"PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#> " + 
				"PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> " + 
				"SELECT ?QualityView ?ScreeningVariable ?dim ?min ?max " +
				"WHERE { " +
				"?QualityView rdf:type dqssv:QualityView; " +
				"dqssv:hasDataVariable dqssv:DATAVARIABLE; " +
				"dqssv:hasScreeningVariable ?ScreeningVariable . " +
				"?ScreeningVariable dqssv:hasDimensionLimits ?dim . " +
				"?dim dqssv:hasMinimumIndex ?min ; " +
				"dqssv:hasMaximumIndex ?max }" ;
			
    	   			     
    	 }
    	 else if (dataVariable.startsWith("TAirStd")) {
    		
    		
    		 this.Query = "PREFIX dqssv: <http://org.esipfed.dataquality/dqss#> " +
    				"PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#> " + 
    				"PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> " + 
    				"SELECT ?QualityView ?ScreeningVariable " +
    				"WHERE { " +
    				  "?QualityView dqssv:hasDataVariable dqssv:DATAVARIABLE; " +
    			     "rdf:type dqssv:QualityView; " +
    	    	     "dqssv:hasScreeningVariable ?ScreeningVariable ." +
    	    	     "?ScreeningVariable dqssv:hasIntegerQualityLevel CRITERIA}";
    		 
         }  
    	 else if (dataVariable.startsWith("CH4_total")) {
    		 
    		 this.Query = "PREFIX dqssv: <http://org.esipfed.dataquality/dqss#> " +
				"PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#> " + 
				"PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> " + 
				"SELECT ?QualityView ?ScreeningVariable " +
				"WHERE { " +
				  "?QualityView dqssv:hasDataVariable dqssv:DATAVARIABLE; " +
			     "rdf:type dqssv:QualityView; " +
	    	     "dqssv:hasScreeningVariable ?ScreeningVariable } ";
	    	     
    	 }
    	 else {
    		 Masker.ReportError("Could not find query"  , 0);
    		 Status = 1;
    	 }
     }
     public Document getResults()
     {
    	 //String qs = getScreeningVariableFromOntologyResults(item);
    	 return this.Results;
     }
     private String getScreeningVariableFromOntologyResults(String item)
     {
     	XPath xpathSelector = DocumentHelper.createXPath("//" + item);
 		List results = xpathSelector.selectNodes(Results);
     	String qs = null;
     	for (Iterator iter = results.iterator(); iter.hasNext(); ) {
 			Element tag = (Element)iter.next();
 			  qs = tag.getStringValue();
 			    	  		
 		}
     	Masker.ReportError("SQDS Screening variable is:" + qs,2);
     	return qs;
     }
     private void fillQuery(String dataVariable, String criteria) {
    	 this.Query = this.Query.replaceAll("CRITERIA" , criteria);
    	 this.Query = this.Query.replaceAll("DATAVARIABLE", dataVariable);
    	 Masker.ReportError(this.Query,2);
     }
     
     
     private void doQuery() {


       Query query = QueryFactory.create(this.Query);
       System.out.println(query.toString());
       // Kind of like an sth for for DBI.  Set up your query and store it in a
       // QueryExecution objects
       QueryExecution qexec = QueryExecutionFactory.create(this.Query, this.dqss);
       
       System.out.println("Printing results of query:" + qexec.toString());
    
 	   try {
    	   
			   
			Element root = Results.addElement("root");
  		    
		
           /* Query Execution
           * When executing the query you must choose what type of query you are
           * executing.  The Javadoc lists the different types, in my case we
           * are using an execSelect() method for the Select query
           */

           /*This example was hard to find
            * here is the local link in the Jena documentation
            * http://jena.sourceforge.net/ARQ/app_api.html
            */
            ResultSet results = qexec.execSelect() ;
            
            System.out.println(results.toString());
            System.out.println(results.getResultVars());
            Element data   = root.addElement("QualityView");
            //Element screen = root.addElement("ScreeningVariable");
            List titles = results.getResultVars();
            
            for (Iterator title = titles.iterator(); title.hasNext(); ) {
            	String x = (String) title.next();
            	System.out.println("LIST:" + x);
            	
            }
            int count = 0;
      
            HashMap<String,HashMap<String,String>> valueMap = new HashMap<String, HashMap<String,String>>();
            HashMap<String,String> countMap = null;

            for (;results.hasNext();)
            {
                 QuerySolution soln = results.nextSolution() ;

                 String key ="";
                 countMap = new HashMap<String, String>();
                 String current = null;
                 for (Iterator<String> it = titles.iterator();it.hasNext();)
                         {

                   String name = it.next();
                   RDFNode x = soln.get(name) ;
                   //System.out.println(x.toString());
                   if(x.isResource())
                   {
                           String subSt[] = x.toString().split("#");
                           System.out.println(name + " -> " + subSt[1]);
                           countMap.put(name,subSt[1]);
                           Element n = root.addElement(name);
                           n.setText(subSt[1]);
                           current = subSt[1];
                   }
                   else
                   {
                	 
                	String subSt[] = x.toString().split("http");
                	System.out.println("Not a resource:" + subSt[0]);
                   countMap.put(name,subSt[0]);
                   Element v = root.addElement(current);
                   v.setText(subSt[0]);
                   }

                 }
                 key = new Integer(count).toString() ;

                 valueMap.put(key, countMap);
                 count++;
            }

            Collection c = valueMap.values();
            
            //obtain an Iterator for Collection
            Iterator itr = c.iterator();
         
            //iterate through HashMap values iterator
            while(itr.hasNext()) {
            	HashMap<String,String> inner = (HashMap<String,String>) itr.next();
            	Collection d = inner.values();
            	
            	Iterator inneriter = d.iterator();
            	while (inneriter.hasNext()) {
            		System.out.println("ITEMS:" + inneriter.next());
            	}
          }
          
     
     /*
             for ( ; results.hasNext() ; )
            {
            	 
                 QuerySolution soln = results.nextSolution() ;
                 
                 RDFNode x = soln.get("QualityView") ;       // Get a result variable by name.
                 RDFNode y = soln.get("ScreeningVariable") ;       // Get a result variable by name.
                 //this.Results = this.Results + (x + " " + y);
                 String[] qv = x.toString().split("#");
                 String[] sv = y.toString().split("#");
                 data.setText(qv[1]);
                 screen.setText(sv[1]);
                 System.out.println(x.toString() + "\n"+ y.toString());
                 
                 
                 }
                 */
        }
       finally { 
     	  qexec.close() ; 
        }
      
 	}
    
}
