package gov.nasa.gsfc.dqss.test;

import gov.nasa.gsfc.dqss.ontology_stub.Ontology;

import java.io.File;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import ncsa.hdf.object.Dataset;
import ncsa.hdf.object.Datatype;
import ncsa.hdf.object.FileFormat;
import ncsa.hdf.object.Group;
import ncsa.hdf.object.HObject;
import ncsa.hdf.object.h4.H4Datatype;
import ncsa.hdf.object.h4.H4File;

import org.dom4j.Document;
import org.dom4j.DocumentHelper;
import org.dom4j.Element;
import org.dom4j.XPath;
/**
 * <p>Title: Reading HDF4 Files with JAVA HDF4 Dataset class</p>
 * <p>Description: This class will:
 * 1. Read in a file.
 * 2. Read in 3 parameters
 * 3. Call the visualizer
 * 4. Apply a simple quality dataset to it's parameter
 * 5. Write out the new datasets
 * 6. Insert the visualized jpeg.
 *  
 * <pre>
 *     "not sure what I should put here yet.
 * </pre>
 * </p>
 *
 * @author Richard Strub GES-DISC
 * @version 1.0
 */
public class HDF4_Read 
{
    private  String fname = "H4FileStructure.hdf";
    private String  inputFile = "/var/scratch/rstrub/AIRS.2010.02.01.042.L2.RetStd.v5.2.2.0.G10033064649.hdf";
    
    
   public HDF4_Read(String param,  String criteria, String inputFile)
   {
       
   	File input = new File(inputFile);
       System.out.println("Opening: "+inputFile);
      
       H4File fileobj = new H4File(input.getPath());
       FileFormat DataFile = null;
       FileFormat fileFormat = FileFormat.getFileFormat(FileFormat.FILE_TYPE_HDF4);
       
       H4File newFile = null;
       
       if (fileobj == null)
       {
           System.err.println("Cannot find HDF4 FileFormat.");
           return;
       }
       
       try {
       	newFile = (H4File)fileFormat.create(fname);
       	System.out.println("Fileformat from: "+inputFile);
       	DataFile = fileobj.open(inputFile, FileFormat.WRITE);

       	if (DataFile == null)
       	{
       		System.err.println("Failed to open file: "+inputFile);
       		return;
       	}

       	// open the file and retrieve the file structure
       	DataFile.open();
       }
       catch (Exception e) {
       	System.out.println("Problem with:" + inputFile);
       }
       
       Group TopGroup = (Group)((javax.swing.tree.DefaultMutableTreeNode)DataFile.getRootNode()).getUserObject();
      
       
   	   determineAndPerformAlgorithm(param,"nBestStd",TopGroup,newFile);
       /*
       Ontology ont = new Ontology("AIRX2STD", tairstd, criteria);
       Document doc = ont.getResults();
       ReportError(doc.asXML(),2);
       if (doc != null) {
       
           


       	HDF4_Read inst = new HDF4_Read();
       	String qs = inst.getScreeningVariableFromOntologyResults("ScreeningVariable",doc);
       	inst.determineAndPerformAlgorithm(tairstd,qs,TopGroup,newFile);
       }
       */
       	/*
       	try {
       		Dataset d = (Dataset) fileobj.get("/L2_Standard_atmospheric&surface_product/SDS Fields/" + qs);
       		ReportError("path:" + d.getName(),2);
       		inst.describeDataset(d);
       	} catch (Exception e1) {
			// TODO Auto-generated catch block
       		e1.printStackTrace();
       	}
       	*/
       	
          	
       // recursively display all of the Datasets of the Groups in the file
       //old printGroup(root,testFile);

       // close file resource
       try {
			DataFile.close();
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

   }
    
    
    public static void main( String args[] ) 
    {
    	String tairstd = args[1];
        String criteria = args[2];
        String inputFile = args[0];
        
        HDF4_Read inst = new HDF4_Read(tairstd,criteria,inputFile);
 } // end main
    private String getScreeningVariableFromOntologyResults(String item, Document doc)
    {
    	XPath xpathSelector = DocumentHelper.createXPath("//" + item);
		List results = xpathSelector.selectNodes(doc);
    	String qs = null;
    	for (Iterator iter = results.iterator(); iter.hasNext(); ) {
			Element tag = (Element)iter.next();
			  qs = tag.getStringValue();
			    	  		
		}
    	ReportError("SQDS Screening variable is:" + qs,2);
    	return qs;
    }
    private void determineAndPerformAlgorithm(String parameter, String qs, Group g, H4File newFile) {
    	String[] qs_vars = new String[1] ;
    	qs_vars[0] = qs;
        Dataset[] qs_var   = getQSVars(parameter,qs_vars,g,newFile);
        if (qs_var[0] != null) {
        	ReportError("I Found qs_var:" + qs_var[0].getName(),2);
        }
        else {
        	ReportError("I could not find a qs_var",2);
        }
        Dataset dtairstd =   dtairstd = getDataset(parameter, g, newFile);
        /* 
         * This is probably where the determineDatatype etc should go.
         * For my first algorithm, I'll hardcode the types and rank.
         */
        LetsSeeIfICanDoSomeArithmeticBetweenDatasetsAlgorithm(dtairstd,qs_var);
        
    }
  
    private void LetsSeeIfICanDoSomeArithmeticBetweenDatasetsAlgorithm(Dataset dtairstd, Dataset[] qs) {
    
    	/*
    	 * Ok, what I am going to do here is to read in TAirStd and nBestStd.
    	 * The Dataset.read() method should read nBestStd in as one array. 
    	 * Whether it is in 2D or not I don't care as long as it reads in each
    	 * TAirStd Layer in the same manner.
    	 * So I will read the first layer.
    	 * For every qs[i] if the layer number < qs[i] I'll set it to 0 otherwise
    	 * I'll set it to 1. 
    	 */
    	Dataset dqs = qs[0];
    	
    	try {
    		
			Dim3SingleQS(dtairstd,dqs,3,45,30,28);
			//displayValues(45,30,28,dtairstd);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
    }
   
    private void Dim3SingleQS(Dataset parameter, Dataset qs, int rank,int rows, int cols, int layers ) throws Exception
    {
    	short[] quality = (short[]) qs.getData();
    	long[] start = parameter.getStartDims(); // get startDims so we can navigate the layers
    	long nlayer = start[2];
    	
    	//First lets see if there is anything interesting in our quality variable
    	for (int i = 0; i < rows * cols;++i) {
    		System.out.print(quality[i] + ",");
    	}
    	//displayValues(45,30,6,parameter);
    	ReportError ("length:" + quality.length,2);
    	float before;
    	for (int layer = 0 ; layer < layers; ++layer) {
    		start[2] = layer; // this sets the layer outside the class as dims are passed by reference.
    		float[] data = (float[]) parameter.read();
    		for (int i = 0; i < rows * cols;++i) {
        		System.out.print(data[i] + "\n" );
        	}
    		for (int i = 0; i < rows * cols; ++i) {
    			 before = data[i];
    			if (layer < quality[i]) {
    		
    				data[i] = 0;
    			}
    			else {
    				data[i] = 1;
    			}
    		    //parameter.write(data);
    			ReportError("layer:"+layer + " quality:" + quality[i] +
    					" Therefore:" + data[i] + " before:" + before,2);
    		}
    		parameter.write(data);
    		
    	}
    	
    }
    
  private Dataset[] getQSVars(String dataVariable, String[] qs,Group g, H4File newFile) {
    	 
    	Dataset[] qsDatasets = new Dataset[qs.length]; 
    	
    	for (int i = 0; i < qs.length; ++i) {
    	    	   ReportError("Doing:" + qs[i],2);
    	   qsDatasets[i] = getDataset(qs[i],g,newFile);
    	   ReportError("getQSVars:" + ((Dataset) qsDatasets[i]).getName(),2);
    	}
    	
    	
    	return (Dataset[]) qsDatasets;
  }
  
  private Dataset createDataset(Group g, H4File newFile, String name, Datatype dtype, long[] dims)
  {
	  Dataset dataset = null;
	try {
		ReportError("Trying to create a new:"+ name,2);
		ReportError(newFile.toString(),2);
		 Datatype ntype = newFile.createDatatype(
		            dtype.getDatatypeClass(), dtype.getDatatypeSize(),dtype.getDatatypeOrder(), dtype.getDatatypeSign());
		 ReportError("created datatype:",2);
		 Group newg = newFile.createGroup(g.getName(), g);
		 ReportError("created group:",2);
		
		dataset = newFile.createScalarDS (name + "_new", newg, ntype, dims, null, null, 0, null);
		ReportError("Created a new:"+ name,2);
		
	} catch (Exception e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
	}
	  return dataset;
  }
  
private Dataset getDataset(String datasetName , Group TopGroup, H4File newFile)
{
	Dataset d = null;
	
	d = getDatasetRecurse(datasetName,TopGroup,d);
	/*Dataset full;
	try {
		full = (Dataset) newFile.get(d.getFullName());
		ReportError(d.getFullName(),2);
		ReportError(full.getName(),2);
		
		describeDataset(full);
	} catch (Exception e1) {
		// TODO Auto-generated catch block
		e1.printStackTrace();
	}
	//d = newFile.get(d.getPath() d.getf)
    */
    
	if (d != null) {
		ReportError("getDataset Found:"+ d.getName(),1);
		Dataset copy  =  createDataset(TopGroup, newFile, d.getName(),d.getDatatype(), d.getDims());
	    try {
			populateDataset(d,copy);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}	
	    return copy;
		}
	else {
		ReportError("getDataset Could not find "+ datasetName,0);
	}
	return d;
}

private void describeDataset(Dataset d)
{
	try {
	Datatype Type = d.getDatatype();
	ReportError("dd:datatype:" + Type.getDatatypeDescription(),2);
	ReportError("dd:name:" + Type.getName(),2);
	}
	catch (NullPointerException e) {
		e.printStackTrace();
		ReportError("dataset was null",2);
	}

}
private void populateDataset(Dataset old, Dataset dnew) throws Exception {
	

		System.out.println("old dataset name:" + old.getName());
		 HashMap dimension = new HashMap();  
         determineDatatype(old,dimension);
		 determineDimensions(old,dimension);
		 
        int[] dims = (int[])dimension.get("dims");
        //int rows = dims[0];    
        //int cols = dims[1];
		 if (dimension.get("type") == "INT16") {
      	      short[] dataRead = (short[])old.read();
                 dnew.write(dataRead);
         }
         if (dimension.get("type") == "INT32") {
      	     int[] dataRead = (int[])old.read();
      	 	dnew.write(dataRead);
         }
             if (dimension.get("type") == "FLOAT32") {
            	 long oldstart[] = old.getStartDims();
            	 long newstart[] = dnew.getStartDims();
            	 for (int i = 0; i < dims[2]; ++i) {
            		 newstart[2] = i;
            		 oldstart[2] = i;
            		 float[] dataRead = (float[])old.read();
                	 dnew.write(dataRead);
            	 }
           	 	 //displayValues(45,30,28,dnew);
             }
             if (dimension.get("type") == "FLOAT64") {
            	 double[] dataRead = (double[])old.read();
           	 	 dnew.write(dataRead);
             }
        
	
}
  // This needs to be recursive as there are groups within groups
private Dataset getDatasetRecurse(String datasetName , Group TopGroup, Dataset d)
{
	List members = TopGroup.getMemberList();

    int n = members.size();
    HObject obj = null;
    int objectid; 
    Dataset looking = null;
  
    for (int i=0; i< members.size(); i++)
    {
    	  if (d != null && d instanceof Dataset) {
    	    	ReportError("Returning:" + d.getName(),2);
    	    	return d;
    	    }
        obj = (HObject)members.get(i);

        /*
         * Each Dataset (SDS) is a member of a Group
         * In the AIRS file I worked with the groups were:
         * L2_Standard_atmospheric&surface_product
         * Geolocation Fields
         * SDS Fields
         * Swath Attributes
         */
        if (obj instanceof Dataset)
        {
        	looking = (Dataset) obj;
        	
        	if (looking.getName().equals(datasetName)){
        		d = looking;
        		ReportError("Found Dataset:" + d.getName(),2);
        		return d;
        	}
        }
        if (obj instanceof Group && d == null)
        {
            d = getDatasetRecurse(datasetName,(Group)obj,d);
            ReportError("In Group" + obj.getName() ,2);
                       
        }
        
        
    }

   return d;
}

public static void ReportError(String Message, int level) {
	if (level == 0) {
		System.err.println(Message);
		System.out.println(Message);
	}
	if (level == 1) {
		System.err.println(Message);
	}
	if (level == 2) {
		System.out.println(Message);
	}
}
    /**
     * Recursively print a group and its members.
     * @throws Exception
     */
    private static void printGroup(Group g, FileFormat testFile) throws Exception
    {
        if (g == null)
            return;

        List members = g.getMemberList();

        int n = members.size();
        HObject obj = null;
        int objectid; 
        for (int i=0; i< members.size(); i++)
        {
            obj = (HObject)members.get(i);
               
            /*
             * Each Dataset (SDS) is a member of a Group
             * In the AIRS file I worked with the groups were:
             * L2_Standard_atmospheric&surface_product
	     * Geolocation Fields
	     * SDS Fields
             * Swath Attributes
             */
            if (obj instanceof Dataset)
            {
               ReadAndDisplaySDS(obj,testFile);
            }
            else {
                System.out.println("Not trying to print Group:"+obj.getName());
            }
            if (obj instanceof Group)
            {
                printGroup((Group)obj,testFile);
            }
        }
    }

    /**
     * create the file and add groups ans dataset into the file,
     * which is the same as javaExample.H4DatasetCreate
     * @see javaExample.H4DatasetCreate
     * @throws Exception
     */

    private static boolean listable(Dataset dataset) 
    {
       if (dataset.hasAttribute()) { 
          return true;
       }
       return false;
    }

    private static void determineDatatype(Dataset dataset, HashMap sizes) throws Exception
    {
            Datatype Type = dataset.getDatatype();
            H4Datatype HType = (H4Datatype)dataset.getDatatype();
            System.out.println("datatype:" + Type.getDatatypeDescription());
            System.out.println("datatype description:" + HType.getDatatypeDescription());
            //System.out.println("dataset type size:" +  HType.getDatatypeSize()); 
            if (HType.isUnsigned()) {
                //System.out.println("dataset " + dataset.getName() + " is unsigned"); 
            }
            else {
                //System.out.println("dataset " + dataset.getName() + " returns that it is NOT unsigned"); 
            }
            sizes.put("size", HType.getDatatypeSize());
            sizes.put("class",  HType.getDatatypeClass());
            sizes.put("name", dataset.getName());
            if (HType.getDatatypeDescription().contains("float") ) {
                 if (HType.getDatatypeSize() == 4) {
                     sizes.put("type", "FLOAT32");
                 }
                 else if (HType.getDatatypeSize() == 8) {
                     sizes.put("type", "FLOAT64");
                 }
                 else {
                     sizes.put("type", "unknown");
                 }
            }
            if (HType.getDatatypeDescription().contains("integer") ) {
                 if (HType.getDatatypeSize() == 2) {
                     sizes.put("type", "INT16");
                 }
                 else if (HType.getDatatypeSize() == 4) {
                     sizes.put("type", "INT32");
                 }
                 else {
                     sizes.put("type", "unknown");
                 }
            }
                      
    }
    private static void determineDimensions(Dataset dataset, HashMap h) {

            long [] dims = (long []) dataset.getDims();
            int[] d = new int[dims.length];

             System.err.print(dataset.getName()); 
             System.err.println("\tlength:" + d.length); 
             System.err.println("\trank:" + dataset.getRank()); 
             System.out.println("\trank:" + dataset.getRank()); 
             System.err.println("\tpath:" + dataset.getPath()); 
             System.out.println("\tlength:" + d.length); 
            for(int i=0 ; i < d.length  ; ++i) {
               if (dims[i] > 0) {
                System.out.print("\tdimensions:" + dims[i]); 
                d[i] = (int)dims[i];
               }
            } 
            System.out.print("\n"); 
            h.put("n",d.length); 
            h.put("dims",d); 

    }

    private static void ReadAndDisplaySDS(HObject obj, FileFormat ff) throws Exception
    {
    
       
        Dataset dataset = (Dataset)obj;
        /*  
         * Determine if dataset is listable 
         */ 
        if (listable(dataset)){
            HashMap dimension = new HashMap();  
            determineDatatype(dataset,dimension);
 
            determineDimensions(dataset,dimension);
 
            int[] dims = (int[])dimension.get("dims");
            int x = dims[0];    
            int y = dims[1];    
            int z = -1;
            if (dims.length > 2) {
                z = dims[2];
            }

            System.out.println("SDS:Name:"+dimension.get("name"));
            System.out.println("Type:"+dimension.get("type"));

               if (dimension.get("type") == "INT16") {
        	   short[] dataRead = (short[])dataset.read();
                   displayValues(x,y,dataRead);
                   displayMinMax(x,y,dataRead);
               }
               if (dimension.get("type") == "INT32") {
        	   int[] dataRead = (int[])dataset.read();
                   displayValues(x,y,dataRead);
                   displayMinMax(x,y,dataRead);
               }
               if (dimension.get("type") == "FLOAT32") {
       
                   if (dims.length == 3) {
                      //display_ncdumpstyle(dataset,x,y,z); 
                       displayValues(x,y,z,dataset);
                       displayMinMax(x,y,z,dataset);
                    }
                    else { 
        	       float[] dataRead = (float[])dataset.read();
                       displayValues(x,y,dataRead);
                       displayMinMax(x,y,dataRead);
                    }
               }
               if (dimension.get("type") == "FLOAT64") {
        	   double[] dataRead = (double[])dataset.read();
                   displayValues(x,y,dataRead);
                   displayMinMax(x,y,dataRead);
               }
          }
    }


     private void display_ncdumpstyle(Dataset dataset, int x, int y, int z) throws Exception
     {
                     int rank = dataset.getRank();   // number of dimension of the dataset
		     long[] dims2 = dataset.getDims(); // the dimension sizes of the dataset
		     long[] selected = dataset.getSelectedDims(); // the selected size of the dataet
		     long[] start = dataset.getStartDims(); // the off set of the selection
		     long[] stride = dataset.getStride(); // the stride of the dataset
		     int[]  selectedIndex = dataset.getSelectedIndex(); // the selected dimensions for display

		     // select dim1 and dim2 as 2D data for display,and slice through dim0
                     start[0] = 0;  
                     start[1] = 0;  
                     start[2] = 2;  
                     System.err.println("start:" + start[0] + " " +start[1] + " " + start[2]);
                     // Dataset.read() reads orthogonally to ncdump. I'll know better when I ask
                     // Young-In about what ncdump.
                     float[][]dataRead = new float[z][x*y];  
                     for (int startdim = 0; startdim < z; ++startdim) {
			    // when dataset.read() is called, the slection above will be used since
			    // the dimension arrays is passed by reference. Changes of these arrays
			    // outside the dataset object directly change the values of these array
			    // in the dataset object.
                         start[2] = startdim;
        	         float[] data = (float[])dataset.read();
                         for (int i = 0; i < data.length; ++i) {
        	           dataRead[startdim][i] = data[i];
                           if (i == 0 ) {
                             System.err.print(data[i]+ ",");
                           }
                         }
                     }
                     for (int c = 0; c < x*y; ++c) {
                       for (int w = 0; w < z; ++w) {
                             System.out.print(dataRead[w][c]+ ",");
                       }
                       System.out.println();
                     }
} 
   private static void displayValues(int x, int y, short[] dataRead)
   {
	   for (int i=0; i<y; i++) {
		System.out.print("\n"+dataRead[i*x]);
		for (int j=1; j<x; j++) {
			System.out.print(", "+dataRead[i*x+j]);
		}
	   }

    }
   private static void displayValues(int x, int y, int[] dataRead)
   {
	   for (int i=0; i<y; i++) {
		System.out.print("\n"+dataRead[i*x]);
		for (int j=1; j<x; j++) {
			System.out.print(", "+dataRead[i*x+j]);
		}
	   }

    }
   private static void olddisplayValues(int x, int y, float[] dataRead)
   {
	   for (int i=0; i<y; i++) {
		System.out.print("\n"+dataRead[i*x]);
		for (int j=1; j<x; j++) {
			System.out.print("; "+dataRead[i*x+j]);
		}
	   }

    }
   private static void displayValues(int x, int y, float[] dataRead)
   {
	   for (int i=0; i< x * y; i++) {
               
                 System.out.print(dataRead[i] + ", ");
		if (i % 5 == 0 ) {
                  System.out.print("\n");
                }
	   }

    }
 
   private static void displayValues(int x, int y, int z, Dataset dataset) throws Exception
   {
	   long[] start = dataset.getStartDims(); // the off set of the selection
           for(int layer = 0; layer < z; ++layer) {
                start[2] = layer; // get to next layer
	   	   		for (int i=0; i < x*y ; i++) {
        	        float[] dataRead = (float[])dataset.read();
        	        if (dataRead[i] > 0) {
                 	    System.out.print(dataRead[i] + ", ");
        	        }
		 	    	if (i % (x*y - 1) == 0 ) {
                  		System.out.print("\n layer " + layer);
                	}
	   			}
           } 

    }
   
   private static void displayValues(int x, int y, double[] dataRead)
   {
           int f = 1;
	   for (int i=0; i< x * y; i++) {
               
                 System.out.print(dataRead[i] + ", ");
		if (f % 4 == 0 ) {
                  System.out.print("\n");
                }
		if ((f) % 30 == 0 ) {
                  System.out.print("\n\n");
                  f = 0;
                }
                ++f;
	   }

    }

   private static void displayMinMax(int x, int y,short[] dataRead) {
           short max = dataRead[0];
           short min  = dataRead[0];
	   for (int i=0; i<y; i++) {
		for (int j=1; j<x; j++) {
                   if (dataRead[i*x+j] < min) {
                       min = dataRead[i*x+j];
                   }
                   if (dataRead[i*x+j] > max) {
                       max = dataRead[i*x+j];
                   }
                }
           }
	System.out.println("Min:"+min+" Max:"+max);

   }
   private static void displayMinMax(int x, int y,int[] dataRead) {
           int max = dataRead[0];
           int min  = dataRead[0];
	   for (int i=0; i<y; i++) {
		for (int j=1; j<x; j++) {
                   if (dataRead[i*x+j] < min) {
                       min = dataRead[i*x+j];
                   }
                   if (dataRead[i*x+j] > max) {
                       max = dataRead[i*x+j];
                   }
                }
           }
	System.out.println("Min:"+min+" Max:"+max);

   }
   private static void displayMinMax(int x, int y,float[] dataRead) {
           float max =  -1000;
           float min  = 1000;
	   for (int i=0; i<y; i++) {
		for (int j=1; j<x; j++) {
                   if (dataRead[i*x+j] < min) {
                       min = dataRead[i*x+j];
                   }
                   if (dataRead[i*x+j] > max) {
                       max = dataRead[i*x+j];
                   }
                }
           }
	System.out.println("Min:"+min+" Max:"+max);

   }
   private static void displayMinMax(int x, int y, int z, Dataset dataset) throws Exception
   {
       float max =  -9999;
       float min  = 10000;
	   long[] start = dataset.getStartDims(); // the off set of the selection
           int f = 1;
           for(int layer = 0; layer < z; ++layer) {
                start[2] = layer; // get to next layer
        	    float[] dataRead = (float[])dataset.read();
	   	   		for (int i=0; i < x*y ; i++) {
                   if (dataRead[i] < min) {
                       min = dataRead[i];
                   }
                   if (dataRead[i] > max) {
                       max = dataRead[i];
                   }
	   			}
           } 
	       System.out.println("Min:"+min+" Max:"+max);

    }

   private static void displayMinMax(int x, int y,double[] dataRead) {
           double max = dataRead[0];
           double min  = dataRead[0];
	   for (int i=0; i<y; i++) {
		for (int j=1; j<x; j++) {
                   if (dataRead[i*x+j] < min) {
                       min = dataRead[i*x+j];
                   }
                   if (dataRead[i*x+j] > max) {
                       max = dataRead[i*x+j];
                   }
                }
           }
	System.out.println("Min:"+min+" Max:"+max);

   }
  
}
